#
# Copyright (c) 2024, HILSTER - https://hilster.io
# All rights reserved.
#

from typing import Callable, Any, Optional, Union
import asyncio
from functools import wraps
import time

from htf.exceptions import SkipTest, ThreadExit
import re
from htf.debug import print_stack_trace
import sys
from htf.environment import get_current_environment

from queue import Queue


def test(func: Callable[..., Any]) -> Callable[..., Any]:
    """
    Mark a function or method as a test.
    """
    func.htf_test = True  # type: ignore
    return func


def periodic(  # noqa: C901: complexity accepted and well tested
    period: float,
    maximum_period: Optional[float] = None,
    run_condition: Optional[Callable[..., bool]] = None,
    raise_exception: bool = True,
) -> Callable[[Callable[..., Any]], Any]:
    """
    Decorate a method to be run periodically.

    Args:
        period: the period in seconds.
        maximum_period=None: if set to a float >= `period`
            the periodic method may take up to `maximum_period` time
            without raising an exception. A warning is printed on
            stdout instead.
        run_condition=None: a callable (method or
            lambda expression) that has to return true while the method
            is run periodically.
        raise_exception=True: if set to ``True`` an exception is raised
            if the called method takes longer
            than the period (or ``maximum_period`` if set).

    Warning:
        Non-realtime operating systems will not ensure timing accuracy.
    """
    if maximum_period is not None:
        if maximum_period < period:
            raise Exception("maximum_period must be >= period!")

    def wrapper(func: Callable[..., Any]) -> Callable[..., Any]:  # noqa: C901: complexity accepted and well tested
        def _check_period(t0: float, t1: float) -> None:
            duration = t1 - t0
            sleep_time = period - duration

            if sleep_time >= 0.0:
                time.sleep(sleep_time)
            else:
                if maximum_period is not None:
                    sleep_time = maximum_period - duration
                    if sleep_time >= 0.0:
                        message = (
                            "Running function in periodic decorator took more than %f seconds (period) "
                            "but took less than %f seconds (maximum_period)! The function took %f seconds."
                            % (period, maximum_period, duration)
                        )
                        print(message)
                    else:
                        message = (
                            "Running function in periodic decorator took more than %f seconds (period) "
                            "and took more than %f seconds (maximum_period)! The function took %f seconds."
                            % (period, maximum_period, duration)
                        )
                        if raise_exception:
                            message += " Cannot continue!"
                            raise Exception(message)
                        else:
                            print(message)
                else:
                    message = (
                        "Running function in periodic decorator took more than %f seconds (period)!"
                        " The function took %f seconds." % (period, duration)
                    )
                    if raise_exception:
                        message += " Cannot continue!"
                        raise Exception(message)
                    else:
                        print(message)

        @wraps(func)
        async def _async_periodic_runner(*args: Any, **kwargs: Any) -> Any:
            if run_condition is not None and not asyncio.iscoroutinefunction(run_condition):
                raise TypeError("run_condition must be a coroutine")

            async def run() -> bool:
                if run_condition is None:
                    return True
                else:
                    return await run_condition()  # type: ignore

            while await run():
                t0 = time.time()
                await func(*args, **kwargs)
                t1 = time.time()
                _check_period(t0, t1)

        @wraps(func)
        def _periodic_runner(*args: Any, **kwargs: Any) -> Any:
            def run() -> bool:
                if run_condition is None:
                    return True
                else:
                    return run_condition()

            while run():
                t0 = time.time()
                func(*args, **kwargs)
                t1 = time.time()
                _check_period(t0, t1)

        if asyncio.iscoroutinefunction(func) or asyncio.iscoroutinefunction(run_condition):
            return _async_periodic_runner
        else:
            return _periodic_runner

    return wrapper


def enqueue_exception(  # noqa: C901: complexity accepted and well tested
    queue: Queue,
) -> Callable[[Callable[..., Any]], Any]:
    """
    Decorate a method to fetch its exception that is enqueued into an
    exception queue.
    The enqueued exception can be handled elsewhere.
    The caught exception can be raised after enqueueing it.

    Just decorate you method with the queue to be used and
    every exception that is raised within the
    decorated method will be put into the queue
    before the exception is raised again.

    Args:
        queue: the queue to enqueue the exception.
    """

    def wrapper(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        async def _async_enqueue_exception(*args: Any, **kwargs: Any) -> None:
            try:
                await func(*args, **kwargs)
            except ThreadExit:  # pragma: no cover
                pass  # ignore ThreadExit in threads, since it means stop() has been called
            except BaseException:
                info = sys.exc_info()
                queue.put(info)
                raise

        def _enqueue_exception(*args: Any, **kwargs: Any) -> None:
            try:
                func(*args, **kwargs)
            except ThreadExit:  # pragma: no cover
                pass  # ignore ThreadExit in threads, since it means stop() has been called
            except BaseException:
                info = sys.exc_info()
                queue.put(info)
                raise

        if asyncio.iscoroutinefunction(func):
            return _async_enqueue_exception
        else:
            return _enqueue_exception

    return wrapper


def timed(limit: float) -> Callable[[Callable[..., Any]], Any]:
    """
    Decorate a method to ensure that it ends within a given time limit.

    Args:
        limit: the time limit in seconds.
    """

    def wrapper(func: Callable[..., Any]) -> Callable[..., Any]:
        def _check_time(t0: float) -> None:
            t1 = time.time()
            td = t1 - t0

            assert limit >= td, "The method did not run within the time limit %f seconds. It took %f seconds." % (
                limit,
                td,
            )

        @wraps(func)
        async def _async_timed(*args: Any, **kwargs: Any) -> Any:
            t0 = time.time()
            await func(*args, **kwargs)
            _check_time(t0)

        @wraps(func)
        def _timed(*args: Any, **kwargs: Any) -> Any:
            t0 = time.time()
            func(*args, **kwargs)
            _check_time(t0)

        if asyncio.iscoroutinefunction(func):
            return _async_timed
        else:
            return _timed

    return wrapper


def raises(*exceptions: BaseException) -> Callable[[Callable[..., Any]], Any]:
    """
    Decorate a method to catch different exceptions.

    Args:
        *exceptions: a tuple of exceptions that are fetched.
    """

    def wrapper(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        async def _async_raises(*args: Any, **kwargs: Any) -> Any:
            try:
                await func(*args, **kwargs)
            except exceptions:  # type: ignore
                pass
            else:
                raise Exception("No exception was raised!")

        def _raises(*args: Any, **kwargs: Any) -> Any:
            try:
                func(*args, **kwargs)
            except exceptions:  # type: ignore
                pass
            else:
                raise Exception("No exception was raised!")

        if asyncio.iscoroutinefunction(func):
            return _async_raises
        else:
            return _raises

    return wrapper


def skip_if(condition: Union[bool, Callable[..., bool]], reason: str) -> Callable[[Callable[..., Any]], Any]:
    """
    Decorate a method to be skipped. The ``condition`` is evaluated at runtime.
    Note that ``condition`` is evaluated twice by the test runner and the test itself. Thus, it should not
    introduce any side effects.

    Args:
        condition: the condition may be ``True``,
            ``False`` or a callable object that must return
            ``True`` or ``False``.
            The decorated test is skipped if ``condition`` is
            ``True`` or returns ``True``.
        reason: the reason to be put into the test report.

    Raises:
        SkipTest: to skip the test if ``condition`` is or returns ``True``.
    """

    def wrapper(func: Callable[..., Any]) -> Callable[..., Any]:
        func.htf_skip_condition = condition  # type: ignore  # this can be evaluated by the test runner to not
        # instantiate fixtures

        @wraps(func)
        async def _skip(*args: Any, **kwargs: Any) -> Any:
            ret = condition
            if hasattr(condition, "__call__"):
                try:
                    if asyncio.iscoroutinefunction(condition):
                        ret = await condition(*args, **kwargs)
                    else:
                        ret = condition(*args, **kwargs)
                except TypeError:
                    if asyncio.iscoroutinefunction(condition):
                        ret = await condition()
                    else:
                        ret = condition()  # type: ignore
            if ret is True:
                raise SkipTest(reason)
            if asyncio.iscoroutinefunction(func):
                await func(*args, **kwargs)
            else:
                func(*args, **kwargs)

        return _skip

    return wrapper


def skip(reason: str) -> Callable[[Callable[..., Any]], Any]:
    """
    Decorate a method to be skipped in view of the occasion.

    Args:
        reason: the reason to be put into the test report.

    Raises:
        SkipTest: to skip the test.
    """
    return skip_if(True, reason)


def meets_doors_requirements(base_url: str, *requirements: str) -> Callable[[Callable[..., Any]], Any]:
    """
    The ``@meets_doors_requirements`` decorator reports which requirements
    are checked by a test in the test report
    shown in the :mod:`~htf.report.doors_testreport`
    and the :mod:`~htf.report.html_testreport.HTMLTestReport`.

    Args:
        base_url: the DOORS base url the links are created with.
        *requirements: a tuple of strings that name the
            checked requirements, eg. "REQ_1", "REQ_2", etc.
    """

    def wrapper(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        async def _links_wrapper(*args: Any, **kwargs: Any) -> Any:
            environment = get_current_environment()
            result = await environment.get_fixture("result")  # the current result fixture

            # the link must be a dictionary with the keys 'title' and 'href'.
            links = []

            for requirement in requirements:
                match = re.match(r"^.*(\d+)", requirement)
                href = "#"
                if match:
                    the_id = match.groups()[0]
                    href = base_url + the_id
                links.append({"title": requirement, "href": href})

            result.links += links

            if asyncio.iscoroutinefunction(func):
                await func(*args, **kwargs)
            else:
                func(*args, **kwargs)

        return _links_wrapper

    return wrapper


def requirements(*requirements: str) -> Callable[[Callable[..., Any]], Any]:
    """
    The ``@requirements`` decorator reports which requirements
    are checked by a test in the test report
    shown in the :mod:`~htf.report.html_testreport.HTMLTestReport`.

    Args:
        *requirements: a tuple of strings that name the
            checked requirements, eg. "REQ_1", "REQ_2", etc.
    """

    def wrapper(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        async def _requirements_wrapper(*args: Any, **kwargs: Any) -> Any:
            environment = get_current_environment()
            result = await environment.get_fixture("result")  # the current result fixture

            result.requirements += [str(requirement) for requirement in requirements]

            if asyncio.iscoroutinefunction(func):
                await func(*args, **kwargs)
            else:
                func(*args, **kwargs)

        return _requirements_wrapper

    return wrapper


def stacktrace(func: Callable[..., Any]) -> Callable[..., Any]:
    """
    The ``@stacktrace`` decorator decorates a method so that its call stack
    is printed to stdout every time it is called.
    This is useful for debugging.
    """

    @wraps(func)
    async def _async_stacktrace(*args: Any, **kwargs: Any) -> Any:
        print_stack_trace(func.__name__)
        return await func(*args, **kwargs)

    @wraps(func)
    def _stacktrace(*args: Any, **kwargs: Any) -> Any:
        print_stack_trace(func.__name__)
        return func(*args, **kwargs)

    if asyncio.iscoroutinefunction(func):
        return _async_stacktrace
    else:
        return _stacktrace


def before(
    before_function: Callable[..., Any], *before_args: Any, **before_kwargs: Any
) -> Callable[[Callable[..., Any]], Any]:
    """
    The `@before` decorator decorates a method so that `before_function`
    is called before the decorated function is called itself.

    Args:
        before_function: the function to be run before
            the decorated function.
        *before_args: positional arguments to be passed
            to ``beforeFunction``.
        **before_kwargs: keyword-arguments to be passed
            to ``beforeFunction``.
    """

    def wrapper(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        async def _before(*args: Any, **kwargs: Any) -> Any:
            if asyncio.iscoroutinefunction(before_function):
                await before_function(*before_args, **before_kwargs)
            else:
                before_function(*before_args, **before_kwargs)

            if asyncio.iscoroutinefunction(func):
                await func(*args, **kwargs)
            else:
                func(*args, **kwargs)

        return _before

    return wrapper


def after(
    after_function: Callable[..., Any], *after_args: Any, **after_kwargs: Any
) -> Callable[[Callable[..., Any]], Any]:
    """
    The `@after` decorator decorates a method so that `after_function`
    is called after the decorated function is called itself.

    Args:
        after_function: the function to be run after
            the decorated function.
        *after_args: positional arguments to be passed
            to ``afterFunction``.
        **after_kwargs: keyword-arguments to be passed
            to ``afterFunction``.
    """

    def wrapper(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        async def _after(*args: Any, **kwargs: Any) -> Any:
            if asyncio.iscoroutinefunction(func):
                await func(*args, **kwargs)
            else:
                func(*args, **kwargs)

            if asyncio.iscoroutinefunction(after_function):
                await after_function(*after_args, **after_kwargs)
            else:
                after_function(*after_args, **after_kwargs)

        return _after

    return wrapper


def _is_cython(fn: Callable[..., Any]) -> bool:
    """
    Returns whether this function is a cythonized function.

    Args:
        fn: function object

    Returns:
        bool: ``True`` if ``fn`` is a cythonized function else ``False``
    """

    try:  # pragma: no cover
        if hasattr(fn, "__func__"):
            fn = fn.__func__  # Class method, static method
    except KeyError:  # pragma: no cover
        pass

    name = type(fn).__name__
    return name == "cython_function_or_method"
