#
# Copyright (c) 2024, HILSTER - https://hilster.io
# All rights reserved.
#

from typing import Any, List, Union
from types import ModuleType
import fnmatch
import inspect
import itertools
import os
import sys
import traceback
from pydoc import locate
from functools import reduce
import glob

from htf.core import Test, TestType, GherkinScenarioTest, GherkinScenarioOutlineTest
from htf.decorators import _is_cython
from htf.gherkin.feature import Feature
from htf.gherkin.scenario import Scenario
from htf.gherkin.scenario_outline import ScenarioOutline
from htf.termcolors import system
from htf.gherkin.gherkin_steps import GherkinSteps
from htf.gherkin.feature_parser import GherkinFeatureParser


import importlib.util


def load_module(filename: str) -> ModuleType:
    for search_path in [os.getcwd()] + sys.path:  # pragma: no branch
        if filename.startswith(search_path):
            # use the relative path to the found search path to be used as the module's import name
            module_path = os.path.relpath(filename, search_path)
            break
    else:
        # fallback and only use the filename as the module name
        module_path = os.path.splitext(os.path.basename(filename))[0]  # pragma: no cover

    # replace the path separator with a .
    module_name = os.path.splitext(module_path)[0].replace(os.path.sep, ".")

    # load the module
    spec = importlib.util.spec_from_file_location(module_name, filename)
    module = importlib.util.module_from_spec(spec)  # type: ignore
    sys.modules[module_name] = module
    try:
        spec.loader.exec_module(module)  # type: ignore
    except Exception as exc:  # pragma: no cover
        exc.__cause__ = None
        raise DiscoveryError(
            'Loading module "{}" failed with the following error:\n\n'
            "{}".format(filename, "".join(traceback.format_exception(*sys.exc_info())))
        )
    return module


class DiscoveryError(Exception):
    def __init__(self, message: str) -> None:
        super(DiscoveryError, self).__init__(message)
        self.__cause__ = None  # disable exception chaining in Python 3


class Discovery:
    """
    A class responsible for resolving given specifiers into corresponding test and fixture objects.
    """

    def __init__(
        self,
        test_specifiers: List[Union[str, ModuleType]],
        fixture_specifiers: List[Union[str, ModuleType]],
        gherkin_specifiers: List[str],
        file_pattern: str = "test*.py",
    ) -> None:
        """
        Args:
            test_specifiers: strings specifying tests or test locations
            fixture_specifiers:  strings specifying fixtures or fixture locations
            gherkin_specifiers:  strings specifying gherkin statements or their locations
            file_pattern: pattern to use to discover files in given folders
        """
        self._test_specifiers = test_specifiers
        self._fixture_specifiers = fixture_specifiers
        self._gherkin_specifiers = gherkin_specifiers
        self._file_pattern = file_pattern

    @staticmethod
    def _get_modules_from_file_specifier(file: str) -> List[ModuleType]:
        """
        Return a list of modules given a file specifier. The file specifier can contain wildcards.

        Args:
            file: file specifier (can be a wildcard)

        Returns:
            List of module objects
        """
        modules = []
        path = os.path.abspath(file)
        files = glob.glob(path)
        for file in files:
            modules.append(load_module(file))
        return modules

    def _resolve_specifiers(  # noqa: C901: complexity accepted and well tested
        self,
        specifiers: List[str],
        do_print: bool = True,
        filter_files: bool = False,
    ) -> List[Any]:
        """
        Resolves specifiers into modules, classes or functions depending on the specifier.

        Args:
            specifiers(list of str): specifiers to resolve

        Returns:
            List of modules, classes and functions
        """
        resolved_specifiers: List[Any] = []
        for specifier in specifiers:
            if do_print is True:
                print(system(specifier))

            if isinstance(specifier, (type(""), bytes)):
                try:
                    if os.path.isdir(specifier):
                        # handle directory specifiers
                        for root, dirs, files in os.walk(specifier, topdown=False):
                            for name in files:
                                if name.lower().endswith(".feature"):
                                    resolved_specifiers.append(os.path.join(root, name))
                        resolved_specifiers += self._get_modules_from_directory(specifier, filter_files)
                        continue
                except TypeError:  # pragma: no cover
                    raise
                    pass

                try:
                    # handle objects
                    obj = locate(specifier)
                except Exception as exc:
                    exc.__cause__ = None  # disable exception chaining of the pydoc error
                    raise DiscoveryError(
                        'Locating specifier "{}" failed with the following error:\n\n'
                        "{}".format(specifier, "".join(traceback.format_exception(*sys.exc_info())))
                    )
                if obj is not None:
                    resolved_specifiers.append(obj)
                    continue

                try:
                    # handle files and wildcards
                    if os.path.isfile(specifier) or "*" in specifier:
                        if specifier.lower().endswith(".feature"):
                            # assert False, specifier
                            resolved_specifiers += GherkinDiscoveryUtils.get_features_from_file_specifier(specifier)
                            continue
                        resolved_specifiers += self._get_modules_from_file_specifier(specifier)
                        continue
                except TypeError:  # pragma: no cover
                    pass

                raise DiscoveryError("Could not resolve specifier: {}".format(specifier))

            resolved_specifiers.append(specifier)
        return resolved_specifiers

    def _get_modules_from_directory(
        self,
        directory: str,
        filter_files: bool,
    ) -> List[ModuleType]:
        """
        Return a list of modules for a given directory.

        Args:
            directory: path to the directory
            filter_files: if set to ``True`` files are filtered

        Returns:
            List of module objects
        """
        modules = []
        path = os.path.abspath(str(directory))
        files = []
        for root, dirnames, filenames in os.walk(path):
            if filter_files:
                for filename in fnmatch.filter(filenames, self._file_pattern):
                    files.append(os.path.join(root, filename))
            else:
                for filename in fnmatch.filter(filenames, "*.py"):
                    files.append(os.path.join(root, filename))

        for file in files:
            modules.append(load_module(file))
        return modules

    @staticmethod
    def _add_cwd_to_path() -> None:
        cwd = os.getcwd()
        if cwd not in sys.path:
            sys.path.insert(0, cwd)

    def discover_gherkin_statements(self) -> GherkinSteps:
        """
        Discover gherkin statements

        Returns:
            FunctionDictionaries object
        """
        self._add_cwd_to_path()
        gherkin_steps = GherkinSteps()

        for specifier in self._resolve_specifiers(self._gherkin_specifiers, do_print=False):
            if inspect.ismodule(specifier):
                gherkin_steps.combine(self.get_gherkin_statements_from_module(specifier))

            elif inspect.isclass(specifier) or inspect.isfunction(specifier):
                gherkin_steps.combine(GherkinDiscoveryUtils.get_gherkin_statements_from_object(specifier))

        return gherkin_steps

    def discover_tests(self) -> List[TestType]:
        """
        Discover test functions using the test specifiers provided.

        Returns:
            List of test functions
        """

        self._add_cwd_to_path()
        gherkin_steps = self.discover_gherkin_statements()
        tests = []
        for specifier in self._resolve_specifiers(self._test_specifiers, filter_files=True):  # type: ignore
            if inspect.ismodule(specifier):
                tests += self.get_tests_from_module(specifier)

            elif TestDiscoveryUtils.is_test(specifier):
                tests.append(Test(specifier))

            elif isinstance(specifier, str) and os.path.isfile(specifier):  # is feature file
                feature = GherkinFeatureParser(filename=specifier).get_feature_from_file()
                tests.extend(GherkinDiscoveryUtils.create_tests_from_feature(feature, gherkin_steps))  # type: ignore
            else:
                tests += TestDiscoveryUtils.get_tests_from_object(specifier)

        return tests

    def discover_fixtures(self) -> List[Any]:
        """
        Discover fixtures using the fixture specifiers provided.

        Returns:
            List of fixtures
        """
        self._add_cwd_to_path()
        fixtures = []

        for specifier in self._resolve_specifiers(self._fixture_specifiers, do_print=False):  # type: ignore
            if inspect.ismodule(specifier):
                fixtures += list(set(self.get_fixtures_from_module(specifier)))

            elif inspect.isclass(specifier) or inspect.isfunction(specifier):
                fixtures += FixtureDiscoveryUtils.get_fixtures_from_object(specifier)

        return fixtures

    def get_tests_from_module(self, module: ModuleType) -> List[TestType]:
        """
        Return a list of test functions for a given module.

        Args:
            module: module to look for tests in

        Returns:
            List of test functions
        """
        tests = []
        if hasattr(module, "__path__"):  # module is a package
            for m in self._get_modules_from_directory(module.__path__[0], filter_files=True):
                tests += self.get_tests_from_module(m)
        tests += list(
            itertools.chain(
                *map(
                    TestDiscoveryUtils.get_tests_from_object,
                    [val for name, val in inspect.getmembers(module, predicate=inspect.isclass)],
                )
            )
        )
        tests += TestDiscoveryUtils.get_tests_from_object(module)
        return tests

    def get_fixtures_from_module(self, module: ModuleType) -> List[Any]:
        """
        Return a list of fixtures for a given module.

        Args:
            module: module to look for fixtures in

        Returns:
            List of fixtures
        """
        fixtures = []
        if hasattr(module, "__path__"):  # module is a package
            for m in self._get_modules_from_directory(module.__path__[0], filter_files=False):
                fixtures += self.get_fixtures_from_module(m)
        fixtures += list(
            itertools.chain(
                *map(
                    FixtureDiscoveryUtils.get_fixtures_from_object,
                    [val for name, val in inspect.getmembers(module, predicate=inspect.isclass)],
                )
            )
        )
        fixtures += list(
            itertools.chain(
                *map(
                    FixtureDiscoveryUtils.get_fixtures_from_object,
                    [val for name, val in inspect.getmembers(module, predicate=inspect.isfunction)],
                )
            )
        )
        return fixtures

    def get_gherkin_statements_from_module(self, module: ModuleType) -> GherkinSteps:
        """
        Returns a dictionary of gherkin statements capture the gherkin functions the module.

        Args:
            module: module to search inside

        Returns:
            Dictionary of gherkin statements
        """
        function_dictionaries = GherkinSteps()
        if hasattr(module, "__path__"):  # module is a package
            for m in self._get_modules_from_directory(module.__path__[0], filter_files=False):
                function_dictionaries.combine(self.get_gherkin_statements_from_module(m))

        def dict_combiner(a: GherkinSteps, b: GherkinSteps) -> GherkinSteps:
            assert isinstance(a, GherkinSteps)
            assert isinstance(b, GherkinSteps)
            a.combine(b)
            return a

        return reduce(
            dict_combiner,
            [
                o
                for o in map(
                    GherkinDiscoveryUtils.get_gherkin_statements_from_object,
                    [val for name, val in inspect.getmembers(module, predicate=inspect.isfunction)],
                )
            ],
            function_dictionaries,
        )


class TestDiscoveryUtils:
    @staticmethod
    def is_test(obj: Any) -> bool:
        """
        Checks if a given object should be considered a test.

        Args:
            obj: object to inspect

        Returns:
            True if the object is likely a test, False otherwise
        """
        return (
            (inspect.isfunction(obj) or inspect.ismethod(obj) or _is_cython(obj))
            and (obj.__name__.startswith("test") or hasattr(obj, "htf_test"))
            and not hasattr(obj, "htf_fixture")
        )

    @staticmethod
    def get_tests_from_object(obj: Any) -> List[TestType]:
        """
        Returns a list of test functions for a given object. A test function is any function or method whose name
        starts with 'test'.

        Args:
            obj: object to look for tests in

        Returns:
            List of test functions
        """

        return [
            Test(test, obj) for test in [getattr(obj, attr) for attr in dir(obj)] if TestDiscoveryUtils.is_test(test)
        ]


class FixtureDiscoveryUtils:
    @staticmethod
    def is_fixture(obj: Any) -> bool:
        return (
            (inspect.isfunction(obj) or inspect.isclass(obj))
            and hasattr(obj, "htf_fixture")
            and not hasattr(obj, "htf_test")
        )

    @staticmethod
    def get_fixtures_from_object(obj: Any) -> List[Any]:
        """
        Return a list of fixtures for a given object. A fixture is any function or class
        with the 'htf_fixture' attribute.

        Args:
            obj: object to look for fixtures in

        Returns:
            List of fixtures
        """

        return [obj] if FixtureDiscoveryUtils.is_fixture(obj) else []


class GherkinDiscoveryUtils:
    @staticmethod
    def get_gherkin_statements_from_object(obj: Any) -> GherkinSteps:
        """
        Returns a dictionary of gherkin statements that capture the function.

        Args:
            obj: object to look for statements in

        Returns:
            a dictionary of gherkin statements that capture the function
        """
        gherkin_steps = GherkinSteps()

        # check for feature
        if hasattr(obj, "_htf_feature"):
            gherkin_steps.register_feature(obj)

        # check for scenario
        if hasattr(obj, "_htf_scenario"):
            gherkin_steps.register_scenario(obj)

        # check for given
        if hasattr(obj, "_htf_given"):
            gherkin_steps.register_given_precondition(obj)

        # check for when
        if hasattr(obj, "_htf_when"):
            gherkin_steps.register_when_condition(obj)

        # check for then
        if hasattr(obj, "_htf_then"):
            gherkin_steps.register_then_assertion(obj)

        # check for transformation
        if hasattr(obj, "_htf_transformation"):
            gherkin_steps.register_transformation(obj)

        # check for data generator
        if hasattr(obj, "_htf_data_generators"):
            gherkin_steps.register_data_generator(obj)

        return gherkin_steps

    @staticmethod
    def get_features_from_file_specifier(pattern: str) -> List[str]:
        """
        Return a list of features filenames. The file specifier can contain wildcards.

        Args:
            pattern: file specifier (can be a wildcard)

        Returns:
            List of filenames
        """
        files = glob.glob(pattern)
        return list(files)

    @staticmethod
    def create_tests_from_feature(feature: Feature, gherkin_steps: GherkinSteps) -> List[TestType]:
        return [  # type: ignore
            GherkinDiscoveryUtils.create_test_from_scenario(feature, scenario, gherkin_steps)
            for scenario in feature.scenarios
        ] + [
            GherkinDiscoveryUtils.create_data_driven_test_from_scenario_outline(
                feature, scenario_outline, gherkin_steps
            )
            for scenario_outline in feature.scenario_outlines
        ]

    @staticmethod
    def create_test_from_scenario(
        feature: Feature, scenario: Scenario, gherkin_steps: GherkinSteps
    ) -> GherkinScenarioTest:
        return GherkinScenarioTest(feature, scenario, gherkin_steps)

    @staticmethod
    def create_data_driven_test_from_scenario_outline(
        feature: Feature, scenario_outline: ScenarioOutline, gherkin_steps: GherkinSteps
    ) -> GherkinScenarioOutlineTest:
        return GherkinScenarioOutlineTest(feature, scenario_outline, gherkin_steps)
