#
# Copyright (c) 2024, HILSTER - https://hilster.io
# All rights reserved.
#

from typing import Optional, List, Any, Dict, Set
import asyncio
import inspect
import sys
import types
from collections import OrderedDict

from htf.event import EventManager, Event

_current_environment: "TestEnvironment" = None  # type: ignore  # access the current environment


def get_current_environment() -> "TestEnvironment":
    global _current_environment
    return _current_environment


async def get_fixture(name: str) -> Any:
    """
    Returns a fixture object as identified by the name parameter.
    If a fixture requires other fixtures, they will be instantiated, if necessary.
    This is the async variant which returns a coroutine, so ideal for async fixtures.

    Args:
        name: name of the requested fixture

    Returns:
        fixture instance

    Raises:
        Exception: if a circular dependency is detected during resolution
    """

    return await get_current_environment().get_fixture(name)


def get_fixture_sync(name: str) -> Any:
    """
    Returns a fixture object as identified by the name parameter.
    If a fixture requires other fixtures, they will be instantiated, if necessary.
    This is the sync variant which returns the fixture instance.

    Args:
        name: name of the requested fixture

    Returns:
        fixture instance

    Raises:
        Exception: if a circular dependency is detected during resolution
        TypeError: if the instantiated fixture is a coroutine
    """

    return get_current_environment().get_fixture_sync(name)


class TestEnvironment:
    """
    A TestEnvironment manages test fixtures and their life-times depending on received events.
    """

    def __init__(self) -> None:
        """ """
        self._fixtures: Dict[str, Any] = OrderedDict()  # fixtures are generator functions and yielding a fixture object
        self._fixture_objects: Dict[str, Any] = {}  # objects yielded by fixture generators
        self._generator_fixtures: Dict[str, Any] = {}  # generator objects instantiated by calling fixtures
        self._function_fixtures: Dict[str, Any] = {}  # return values of function fixtures
        self._active_scopes: Set[str] = set()  # scopes for which we received '*_started' events, but not '*_finished'
        self._fixture_instantiation_orders: Dict[str, List[str]] = {
            "test": [],
            "class": [],
            "module": [],
            "session": [],
        }  # a dictionary which holds the orders of fixture instantiation for every possible scope to be able to
        # teardown dependant fixtures in the reversed order

        global _current_environment
        _current_environment = self

    async def get_fixture(
        self, name: str, requesting_scope: str = "test", resolving: Optional[List[Any]] = None
    ) -> Any:
        """
        Returns a fixture object as identified by the name parameter. If a fixture requires other fixtures,
        they too will be instantiated if necessary.

        Args:
            name: name of the requested fixture
            requesting_scope: the scope of the requesting fixture
            resolving: an internal parameter used to track the fixture names being actively resolved

        Returns:
            fixture instance

        Raises:
            Exception: if a circular dependency is detected during resolution
        """

        scopes = ["test", "class", "module", "session"]
        assert requesting_scope in scopes, "Possible values for scope: session, module, class, test"

        if resolving is None:
            resolving = []
        if name in resolving:
            raise Exception("Circular dependency in requested fixture: {}.".format(name))
        else:
            resolving.append(name)

        # check if requesting scope is less or equal the fixtures' scope
        try:
            if scopes.index(requesting_scope) > scopes.index(self._fixtures[name]._fixture_scope):
                raise Exception(
                    "Fixture '{}' with scope '{}' cannot be requested in a wider scope '{}'.".format(
                        name, self._fixtures[name]._fixture_scope, requesting_scope
                    )
                )
        except KeyError:
            raise Exception("Could not find fixture: {}.".format(name))

        if name not in self._fixture_objects:
            await self._instantiate_fixture(name, resolving)

        resolving.remove(name)

        return self._fixture_objects[name]

    def get_fixture_sync(self, name: str, requesting_scope: str = "test", resolving: Optional[List[Any]] = None) -> Any:
        """
        This is the sync variant of the 'get_fixture' function.
        Returns a fixture object as identified by the name parameter. If a fixture requires other fixtures,
        they too will be instantiated if necessary.

        Args:
            name: name of the requested fixture
            requesting_scope: the scope of the requesting fixture
            resolving: an internal parameter used to track the fixture names being actively resolved

        Returns:
            fixture instance

        Raises:
            Exception: if a circular dependency is detected during resolution
            TypeError: if the instantiated fixture is a coroutine
        """

        scopes = ["test", "class", "module", "session"]
        assert requesting_scope in scopes, "Possible values for scope: session, module, class, test"

        if resolving is None:
            resolving = []
        if name in resolving:
            raise Exception("Circular dependency in requested fixture: {}.".format(name))
        else:
            resolving.append(name)

        # check if requesting scope is less or equal the fixtures' scope
        try:
            if scopes.index(requesting_scope) > scopes.index(self._fixtures[name]._fixture_scope):
                raise Exception(
                    "Fixture '{}' with scope '{}' cannot be requested in a wider scope '{}'.".format(
                        name, self._fixtures[name]._fixture_scope, requesting_scope
                    )
                )
        except KeyError:
            raise Exception("Could not find fixture: {}.".format(name))

        if name not in self._fixture_objects:
            self._instantiate_fixture_sync(name, resolving)

        resolving.remove(name)

        if inspect.iscoroutine(self._fixture_objects[name]) or inspect.isasyncgen(self._fixture_objects[name]):
            raise TypeError("The fixture must not be an async coroutine")

        return self._fixture_objects[name]

    def get_auto_fixture_sync(self, name: str) -> Any:
        """
        Get an already instantiated fixture in synchronous context.
        For fixtures that are not yet instantiated the 'get_fixture_sync' function shall be used.
        Is used as a helper for htf.TestCase.
        """

        return self._fixture_objects[name]

    def set_fixture(self, name: str, fixture: Any) -> None:
        """
        Registers a fixture with a name it can be requested by.

        Args:
            name: name of the fixture to register
            fixture: fixture generator

        Returns:

        """
        self._fixtures[name] = fixture
        EventManager.register_callback(fixture._fixture_scope + "_started", self.handle_event)
        EventManager.register_callback(fixture._fixture_scope + "_finished", self.handle_event)

    async def _instantiate_fixture(self, name: str, resolving: Optional[List[Any]]) -> None:
        """
        This instantiates the given fixtures and also instantiates all fixtures that might be in the arguments list.
        Args:
            name: the name of the fixture to instantiate.
            resolving: an internal parameter used to track the fixture names being actively resolved

        Returns:

        """

        assert (
            name not in self._fixture_objects and name not in self._generator_fixtures
        ), "The requested fixture ({}) was already instantiated.".format(name)

        scope = self._fixtures[name]._fixture_scope

        if scope in self._active_scopes:
            args = self._get_args(self._fixtures[name])
            if len(args) > 0:
                if asyncio.iscoroutinefunction(self._fixtures[name]):
                    fixture = await self._fixtures[name](
                        *[await self.get_fixture(n, self._fixtures[name]._fixture_scope, resolving) for n in args]
                    )
                else:
                    fixture = self._fixtures[name](
                        *[await self.get_fixture(n, self._fixtures[name]._fixture_scope, resolving) for n in args]
                    )
            else:
                if asyncio.iscoroutinefunction(self._fixtures[name]):
                    fixture = await self._fixtures[name]()
                else:
                    fixture = self._fixtures[name]()

            if isinstance(fixture, types.GeneratorType) or callable(
                getattr(fixture, "__next__", None)
            ):  # sync generator
                # is a generator or mimics the generator protocol
                self._fixture_objects[name] = next(fixture)
                self._generator_fixtures[name] = fixture
                self._fixture_instantiation_orders[scope].append(name)
            elif isinstance(fixture, types.AsyncGeneratorType) or callable(
                getattr(fixture, "__anext__", None)
            ):  # async generator
                self._fixture_objects[name] = await fixture.__anext__()
                self._generator_fixtures[name] = fixture
                self._fixture_instantiation_orders[scope].append(name)
            else:
                # fixture is not a generator
                self._fixture_objects[name] = fixture
                self._function_fixtures[name] = fixture
                self._fixture_instantiation_orders[scope].append(name)
        else:
            raise Exception(
                "Fixture {} was requested outside of its scope '{}'.".format(name, self._fixtures[name]._fixture_scope)
            )

    def _instantiate_fixture_sync(self, name: str, resolving: Optional[List[Any]]) -> None:
        """
        This is the sync variant of the '_instantiate_fixture' function to use only with sync fixtures.
        This instantiates the given fixtures and also instantiates all fixtures that might be in the arguments list.
        Args:
            name: the name of the fixture to instantiate.
            resolving: an internal parameter used to track the fixture names being actively resolved

        Returns:

        """

        assert (
            name not in self._fixture_objects and name not in self._generator_fixtures
        ), "The requested fixture ({}) was already instantiated.".format(name)

        scope = self._fixtures[name]._fixture_scope

        if scope in self._active_scopes:
            args = self._get_args(self._fixtures[name])
            if len(args) > 0:
                fixture = self._fixtures[name](
                    *[self.get_fixture_sync(n, self._fixtures[name]._fixture_scope, resolving) for n in args]
                )
            else:
                fixture = self._fixtures[name]()

            if isinstance(fixture, types.GeneratorType) or callable(
                getattr(fixture, "__next__", None)
            ):  # sync generator
                # is a generator or mimics the generator protocol
                self._fixture_objects[name] = next(fixture)
                self._generator_fixtures[name] = fixture
                self._fixture_instantiation_orders[scope].append(name)
            else:
                # fixture is not a generator
                self._fixture_objects[name] = fixture
                self._function_fixtures[name] = fixture
                self._fixture_instantiation_orders[scope].append(name)
        else:
            raise Exception(
                "Fixture {} was requested outside of its scope '{}'.".format(name, self._fixtures[name]._fixture_scope)
            )

    @staticmethod
    def _get_args(obj: Any) -> List[str]:
        """
        Returns a list of arguments of an object, fixtures in this case. Used to determine if a fixture is
        dependent on another fixture.

        Args:
            obj: object to get arguments for

        Returns:
            List of str of arguments
        """
        args = inspect.getfullargspec(obj).args

        if inspect.isclass(obj):
            args = args[1:]

        return args

    def _clean_up_generator_fixture(self, name: str) -> None:
        """
        Remove any generator fixture objects and active generator objects produced by a fixture.

        Args:
            name: name of the fixture to clean up
        """
        del self._generator_fixtures[name]
        del self._fixture_objects[name]

    def _clean_up_function_fixture(self, name: str) -> None:
        """
        Remove any function fixture objects

        Args:
            name: name of the fixture to clean up
        """
        del self._function_fixtures[name]
        del self._fixture_objects[name]

    async def _process_state_finished(self, scope: str) -> None:  # noqa: C901: complexity accepted and well tested
        """
        When a '*_finished' event is received all fixtures matching the scope are cleaned up if necessary.
        If an exception is raised during the clean up of a fixture it is cached while any other remaining
        fixtures are cleaned up and re-raised after. If multiple exceptions are raised during clean up
        only the exception which occurred first will be re-raised.

        Raises:
            re-raises any exceptions raised during clean up of fixtures
        """
        exception = None

        for name in reversed(self._fixture_instantiation_orders[scope]):
            if self._fixtures[name]._fixture_scope not in self._active_scopes:  # pragma: no branch
                if name in self._generator_fixtures:
                    try:
                        if isinstance(self._generator_fixtures[name], types.GeneratorType) or callable(
                            getattr(self._generator_fixtures[name], "__next__", None)
                        ):  # sync generator
                            next(self._generator_fixtures[name])
                        elif isinstance(self._generator_fixtures[name], types.AsyncGeneratorType) or callable(
                            getattr(self._generator_fixtures[name], "__anext__", None)
                        ):  # async generator  # pragma: no cover
                            await self._generator_fixtures[name].__anext__()
                        raise Exception("Fixture ({}) did not raise StopIteration on tear down.".format(name))
                    except (StopIteration, StopAsyncIteration):
                        self._clean_up_generator_fixture(name)
                    except Exception as ex:
                        self._clean_up_generator_fixture(name)
                        # Python 2.7 doesn't cache exception tracebacks
                        ex._htf_exc_info = sys.exc_info()  # type: ignore
                        exception = exception or ex  # type: ignore
                if name in self._function_fixtures:
                    self._clean_up_function_fixture(name)

        # clear instantiation order for the scope
        self._fixture_instantiation_orders[scope] = []

        if exception is not None:
            raise exception

    async def _process_state_started(self, scope: str) -> None:
        """
        When a '*_started' event is received all fixtures which have the 'auto' attribute set to True are instantiated.
        """
        for name, fixture in list(self._fixtures.items()):
            if fixture._fixture_scope == scope and fixture._fixture_auto is True:
                await self.get_fixture(name, requesting_scope=scope)

    async def _process_state_change(self, scope: str, state: str) -> None:
        """
        Handles scope state changes.
        When a 'started' state is received the scope is added to a set of active scopes.
        On 'finished' states the scope is removed from the set of active scopes and clean up
        is started for any fixtures matching the scope that was left.

        Args:
            scope: the scope which changed state
            state: 'started' or 'finished'

        Raises:
            Exception if a 'finished' state is received for a non-active scope.
        """
        if state == "started":
            self._active_scopes.add(scope)
            await self._process_state_started(scope)
        elif state == "finished":  # pragma: no branch
            try:
                self._active_scopes.remove(scope)
            except KeyError:
                raise Exception("Scope '{}' is not active.".format(scope))
            await self._process_state_finished(scope)

    async def handle_event(self, event: Event) -> None:
        """
        Splits events like 'session_started' or 'test_finished' into the scope and state and processes
        the state change.

        Args:
            event: the event that occurred
        """
        scope, state = event.type.split("_")
        await self._process_state_change(scope, state)

    def get_active_scopes(self) -> Set[str]:
        """
        Returns a set of scopes which are currently active

        Returns:
            set of strings representing active scopes
        """
        return self._active_scopes
