#
# Copyright (c) 2024, HILSTER - https://hilster.io
# All rights reserved.
#

from typing import Type, Optional
from types import TracebackType
import datetime
import sys

from htf.exceptions import SkipTest, NotRunException, ScheduledException, format_exception
from htf.termcolors import failure, error, success, skipped


class Evaluation:
    """
    A context-manager to interpret the result of a test executed in its context.
    """

    def __init__(self, result_type: Type["TestResult"]) -> None:  # type: ignore  # noqa: F821
        """
        Initialize the instance by setting the result type to generate on call.

        Args:
            result_type: the type of result to generate
        """
        self._result_type = result_type
        self.result: TestResult = None  # type: ignore  # noqa: F821
        self.silent = False

    def __enter__(self) -> Optional["TestResult"]:  # type: ignore  # noqa: F821
        """
        Returns:
            instance of the result type
        """
        return self.result

    def __call__(self, test: "Test") -> "Evaluation":  # type: ignore  # noqa: F821
        """
        On call the result type is passed the test it belongs to, and result is generated
        Also enables capture of any output generated with the evaluation context.

        Args:
            test: test function the result belongs to

        Returns:
            the Evaluation instance

        """
        self.result = self._result_type(test)
        self.result.set_output_capture_enable(enabled=True)
        return self

    def __exit__(self, exc_type: Type[BaseException], exc_val: BaseException, exc_tb: TracebackType) -> bool:
        """
        On leaving the context block the result is interpreted by examining the type of exception, if any,
        was raised within the context.

        Args:
            exc_type: exception type
            exc_val: exception value
            exc_tb: exception traceback

        Returns:
            True to mark any exceptions as handled
        """
        self.result.set_exception(exc_type, exc_val, exc_tb)

        message = ""

        if exc_type is None:
            self.result.result = "success"
            message += success("SUCCESS")

        elif exc_type == AssertionError:
            self.result.result = "failure"
            message += failure("FAILURE") + "\n"
            message += failure(
                format_exception(
                    self.result.exception_type,  # type: ignore
                    self.result.exception_value,  # type: ignore
                    self.result.exception_traceback,  # type: ignore
                )
            )

        elif issubclass(exc_type, SkipTest):
            self.result.result = "skipped"
            message += skipped("SKIPPED") + "\n"
            message += skipped(
                format_exception(
                    self.result.exception_type,  # type: ignore
                    self.result.exception_value,  # type: ignore
                    self.result.exception_traceback,  # type: ignore
                )
            )

        elif issubclass(exc_type, NotRunException):
            self.result.result = "skipped"

        elif issubclass(exc_type, ScheduledException):
            self.result.result = "scheduled"

        else:
            self.result.result = "error"
            message += error("ERROR") + "\n"
            message += error(
                format_exception(
                    self.result.exception_type,  # type: ignore
                    self.result.exception_value,  # type: ignore
                    self.result.exception_traceback,  # type: ignore
                )
            )

        if not self.silent and message:
            try:
                print(message)
            except UnicodeEncodeError:  # pragma: no cover
                print(
                    message.encode(encoding=sys.getdefaultencoding(), errors="replace").decode(
                        encoding=sys.getdefaultencoding()
                    )
                )

        self.result.stop_time = datetime.datetime.now()

        self.result.set_output_capture_enable(enabled=False)

        return True
