#
# Copyright (c) 2024, HILSTER - https://hilster.io
# All rights reserved.
#

from typing import Optional, Dict, Callable, List, Any
import asyncio


class Event:
    """
    An event object with a type and optional additional data.
    """

    def __init__(self, event_type: str, kwargs: Optional[Dict] = None) -> None:
        """
        Args:
            event_type: type of the event
            kwargs: optional additional data related to the event.
        """
        self.type = event_type
        self.kwargs = kwargs or {}

    def __str__(self) -> str:
        return "{}: {}".format(self.type, self.kwargs)


class _EventManager:
    """
    A simple EventManager handling callbacks and events.
    """

    def __init__(self) -> None:
        self._callbacks: Dict[str, List[Callable[[Event], Any]]] = {}

    def register_callback(self, event_type: str, callback: Callable[[Event], Any]) -> None:
        """
        Register a callback function for an event type.

        Args:
            event_type: event type to register the callback for
            callback: function to call if a matching event occurs
        """
        if event_type not in self._callbacks:
            self._callbacks[event_type] = []
        if callback not in self._callbacks[event_type]:
            self._callbacks[event_type].append(callback)

    def unregister_callback(self, event_type: str, callback: Callable[[Event], Any]) -> None:
        """
        Remove a callback for an event type.

        Args:
            event_type: event type to unregister the callback for
            callback: function to remove
        """
        try:
            self._callbacks[event_type].remove(callback)
        except ValueError:
            pass

    async def handle_event(self, event: Event) -> None:
        """
        Calls all registered callback functions matching the event until a callback returns True,
        marking the event as handled.

        Args:
            event: event that occurred
        """
        exception = None

        if event.type in self._callbacks:
            for callback in self._callbacks[event.type]:
                try:
                    if asyncio.iscoroutinefunction(callback):
                        if await callback(event):
                            break
                    else:
                        if callback(event):
                            break
                except BaseException as ex:
                    exception = exception or ex  # type: ignore

        if exception is not None:
            raise exception

    def handle_event_sync(self, event: Event) -> None:
        """
        Calls all registered synchronous callback functions matching the event until a callback returns True,
        marking the event as handled.

        Raises an exception if an async callback shall be called.

        Args:
            event: event that occurred
        """
        # print(event)
        if event.type in self._callbacks:
            for callback in self._callbacks[event.type]:  # pragma: no branch
                if asyncio.iscoroutinefunction(callback):
                    raise TypeError(f"A coroutine callback for {event.type} cannot be called from non async context!")
                else:
                    if callback(event):
                        break

    def clear_callbacks(self) -> None:
        """
        Clear all callbacks.
        """
        self._callbacks = {}


EventManager = _EventManager()
