#
# Copyright (c) 2024, HILSTER - https://hilster.io
# All rights reserved.
#

from typing import Generator, Optional, Type
import os
import traceback
from types import TracebackType

import htf.assertions
import htf.core
import warnings
from contextlib import contextmanager

from htf.mixins.delegation import DelegationError  # noqa: F401: needed in tests that use it


class FeatureLocked(Exception):  # noqa: N818
    """
    A requested feature is locked
    """


class SkipTest(Exception):  # noqa: N818
    pass


class SkipStep(SkipTest):
    pass


class RepeatTest(SkipTest):
    pass


class ParsingError(BaseException):
    pass


class NotRunException(Exception):  # noqa: N818
    pass


class ScheduledException(Exception):  # noqa: N818
    pass


class ThreadExit(BaseException):
    pass


class CancelTestException(AssertionError):  # noqa: N818
    pass


def format_exception(
    exception_type: Optional[Type[BaseException]],
    exception_value: Optional[BaseException],
    exception_traceback: Optional[TracebackType],
) -> str:
    try:
        htf_core_path = os.path.dirname(htf.core.__file__)
        htf_assertions_path = htf.assertions.__file__

        tb = traceback.StackSummary.extract(traceback.walk_tb(exception_traceback))

        # remove first frames that point to htf.core
        for start in range(len(tb)):
            if htf_core_path not in tb[start].filename and "htf/core/__init__.pyx" not in tb[start].filename:
                # first stack-frame outside htf.core
                break
        else:
            # break was not run -> keep full traceback!
            start = 0

        # remove first entries
        tb = tb[start:]  # type: ignore

        # remove last frames that lead to htf.assertions
        for stop in range(len(tb) - 1, -1, -1):  # pragma: no branch
            if tb[stop].filename != htf_assertions_path:
                break
        else:  # pragma: no cover
            # break was not run -> keep full traceback!
            stop = len(tb) - 1

        tb = tb[: stop + 1]  # type: ignore

        exc_str = "Traceback (most recent call last)" + "".join(
            [item for item in traceback.StackSummary.from_list(tb).format()]
        )
        exc_str += f"{exception_type.__name__}: {exception_value}"  # type: ignore

        return exc_str
    except Exception as e:
        # in case something goes wrong keep the full traceback
        print(f"WARNING: could not format exception: {e.__class__.__name__}: {str(e)}")
        return "".join(traceback.format_exception(exception_type, exception_value, exception_traceback))


def _format_runtime_warning(warning: warnings.WarningMessage) -> str:
    return f'{warning.message.args[0].capitalize()} (File "{warning.filename}", line {warning.lineno})'  # type: ignore


@contextmanager
def raise_coroutine_runtime_warnings_as_runtime_errors() -> Generator[None, None, None]:
    try:
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always", RuntimeWarning)
            yield
    finally:
        relevant_warnings = [
            _format_runtime_warning(warning)
            for warning in w
            if issubclass(warning.category, RuntimeWarning)
            and "coroutine" in str(warning.message)
            and "was never awaited" in str(warning.message)
        ]

        s = "\n    ".join(relevant_warnings)

        if len(relevant_warnings):
            n = "One" if len(relevant_warnings) == 1 else "Some"
            coro = "coroutine" if len(relevant_warnings) == 1 else "coroutines"
            raise RuntimeError(f"{n} {coro} were never awaited:\n    {s}")
