#
# Copyright (c) 2024, HILSTER - https://hilster.io
# All rights reserved.
#

from typing import Optional, Dict, Any, Callable

from htf.gherkin.step_parsers import get_parser


def feature(
    pattern: str,
    parser: Optional[str] = None,
    extra_types: Optional[Dict[Any, Any]] = None,
    converters: Optional[Dict[str, Any]] = None,
) -> Callable[[Callable[..., Any]], Any]:
    """
    Decorates a method to make it an implementation of a given feature in a behavior driven test.

    It can be implemented as a normal method with a return value to implement
    set up code or as a generator to realize set up and tear down code for a feature.

    Args:
        pattern: gherkin feature pattern
        parser=None: the parser to parse the parameters from the precondition
        extra_types=None: extra types for parsing
        converters=None: optional converters that are applied to
            parameter after parsing
    """
    parser_class = get_parser(parser)

    def wrapper(func: Callable[..., Any]) -> Callable[..., Any]:
        func._htf_feature = True  # type: ignore
        if not hasattr(func, "_htf_features"):  # pragma: no branch
            func._htf_features = set()  # type: ignore

        parser = parser_class(expression=pattern, converters=converters, extra_types=extra_types)
        func._htf_features.add(parser)  # type: ignore

        return func

    return wrapper


def scenario(
    pattern: str,
    parser: Optional[str] = None,
    extra_types: Optional[Dict[Any, Any]] = None,
    converters: Optional[Dict[str, Any]] = None,
) -> Callable[[Callable[..., Any]], Any]:
    """
    Decorates a method to make it an implementation of a given scenario in a behavior driven test.

    It can be implemented as a normal method with a return value to implement
    set up code or as a generator to realize set up and tear down code for a scenario.

    Args:
        pattern: gherkin feature pattern
        parser=None: the parser to parse the parameters from the precondition
        extra_types=None: extra types for parsing
        converters=None: optional converters that are applied to
            parameters after parsing
    """
    parser_class = get_parser(parser)

    def wrapper(func: Callable[..., Any]) -> Callable[..., Any]:
        func._htf_scenario = True  # type: ignore
        if not hasattr(func, "_htf_scenarios"):  # pragma: no branch
            func._htf_scenarios = set()  # type: ignore

        parser = parser_class(expression=pattern, converters=converters, extra_types=extra_types)
        func._htf_scenarios.add(parser)  # type: ignore

        return func

    return wrapper


def given(
    precondition: str,
    parser: Optional[str] = None,
    extra_types: Optional[Dict[Any, Any]] = None,
    converters: Optional[Dict[str, Any]] = None,
) -> Callable[[Callable[..., Any]], Any]:
    """
    Decorates a method to make it an implementation of a given step in a behavior driven test.
    See also https://cucumber.io/docs/gherkin/reference/#given .

    Args:
        precondition: gherkin feature precondition
        parser=None: the parser to parse the parameters from the precondition
        extra_types=None: extra types for parsing
        converters=None: optional converters that are applied to
            parameters after parsing
    """
    parser_class = get_parser(parser)

    def wrapper(func: Callable[..., Any]) -> Callable[..., Any]:
        func._htf_given = True  # type: ignore
        if not hasattr(func, "_htf_given_preconditions"):
            func._htf_given_preconditions = set()  # type: ignore

        parser = parser_class(expression=precondition, converters=converters, extra_types=extra_types)
        func._htf_given_preconditions.add(parser)  # type: ignore

        return func

    return wrapper


def when(
    condition: str,
    parser: Optional[str] = None,
    extra_types: Optional[Dict[Any, Any]] = None,
    converters: Optional[Dict[str, Any]] = None,
) -> Callable[[Callable[..., Any]], Any]:
    """
    Decorates a method to make it an implementation of a when step in a behavior driven test.
    See also https://cucumber.io/docs/gherkin/reference/#when .

    Args:
        condition: gherkin feature condition
        parser=None: the parser to parse the parameters from the condition
        extra_types=None: extra types for parsing
        converters=None: optional converters that are applied to
            parameters after parsing
    """
    parser_class = get_parser(parser)

    def wrapper(func: Callable[..., Any]) -> Callable[..., Any]:
        func._htf_when = True  # type: ignore
        if not hasattr(func, "_htf_when_conditions"):
            func._htf_when_conditions = set()  # type: ignore

        parser = parser_class(expression=condition, converters=converters, extra_types=extra_types)
        func._htf_when_conditions.add(parser)  # type: ignore

        return func

    return wrapper


def then(
    assertion: str,
    parser: Optional[str] = None,
    extra_types: Optional[Dict[Any, Any]] = None,
    converters: Optional[Dict[str, Any]] = None,
) -> Callable[[Callable[..., Any]], Any]:
    """
    Decorates a method to make it an implementation of a then step in a behavior driven test.
    See also https://cucumber.io/docs/gherkin/reference/#then .

    Args:
        assertion: gherkin feature assertion
        parser=None: the parser to parse the parameters from the assertion
        extra_types=None: extra types for conversion
        converters=None: optional converters that are applied to
            parameters after parsing
    """

    parser_class = get_parser(parser)

    def wrapper(func: Callable[..., Any]) -> Callable[..., Any]:
        func._htf_then = True  # type: ignore
        if not hasattr(func, "_htf_then_assertions"):
            func._htf_then_assertions = set()  # type: ignore

        parser = parser_class(expression=assertion, converters=converters, extra_types=extra_types)
        func._htf_then_assertions.add(parser)  # type: ignore

        return func

    return wrapper


def transformation(
    pattern: str,
    parser: Optional[str] = None,
) -> Callable[[Callable[..., Any]], Any]:
    """
    Decorates a method to make it an implementation of a data transformation hook
    used for examples in a behavior driven test.

    Args:
        pattern: gherkin examples pattern
        parser=None: the parser to match the name of the example
    """

    parser_class = get_parser(parser)

    def wrapper(func: Callable[..., Any]) -> Callable[..., Any]:
        func._htf_transformation = True  # type: ignore
        if not hasattr(func, "_htf_transformations"):  # pragma: no branch
            func._htf_transformations = set()  # type: ignore

        parser = parser_class(expression=pattern)
        func._htf_transformations.add(parser)  # type: ignore

        return func

    return wrapper


def data_generator(
    pattern: str,
    parser: Optional[str] = None,
) -> Callable[[Callable[..., Any]], Any]:
    """
    Decorates a method to make it an implementation of a data generator
    used for examples in a behavior driven test.

    Args:
        pattern: the pattern to match the title
        parser=None: the parser to match the name of the example
    """

    parser_class = get_parser(parser)

    def wrapper(func: Callable[..., Any]) -> Callable[..., Any]:
        func._htf_data_generator = True  # type: ignore
        if not hasattr(func, "_htf_data_generators"):  # pragma: no branch
            func._htf_data_generators = set()  # type: ignore

        parser = parser_class(expression=pattern)
        func._htf_data_generators.add(parser)  # type: ignore

        return func

    return wrapper
