#
# Copyright (c) 2024, HILSTER - https://hilster.io
# All rights reserved.
#

from typing import Callable, Any, List, Set, Optional, Tuple, Dict
import inspect
import re

indentation_pattern = re.compile(r"^[ \t]*")


class GherkinSteps:
    """
    This class holds gherkin steps.
    """

    def __init__(self) -> None:
        self._features: Set[Callable[..., Any]] = set()
        self._scenarios: Set[Callable[..., Any]] = set()
        self._given_preconditions: Set[Callable[..., Any]] = set()
        self._when_conditions: Set[Callable[..., Any]] = set()
        self._then_assertions: Set[Callable[..., Any]] = set()
        self._transformations: Set[Callable[..., Any]] = set()
        self._data_generators: Set[Callable[..., Any]] = set()

    @staticmethod
    def get_indentation(line: str) -> int:
        match = indentation_pattern.match(line).group(0)  # type: ignore
        if match:  # pragma: no cover
            return len(match)
        else:  # pragma: no cover
            return 0

    def get_printable_method(self, method: Callable[..., Any]) -> str:
        filename = inspect.getsourcefile(method)
        source, line_number = inspect.getsourcelines(method)

        __str__ = "'{}' defined in File \"{}\", line {}:\n".format(method.__name__, filename, line_number)
        if len(source) > 5:
            last_line = source[5]
            source = source[:5] + [" " * self.get_indentation(last_line) + "# %< %< %< (snipped)"]
        __str__ += "".join(source) + "\n"

        return __str__

    def get_matches_printable(self, matches: List[Tuple[Callable[..., Any], Dict[str, Any]]]) -> str:
        """
        Return all matches in a printable way.
        """
        return "\n".join(self.get_printable_method(match) for match, _ in matches)

    def register_feature(self, implementation: Callable[..., Any]) -> None:
        """
        Register a @feature.

        Args:
            implementation: implementation of a @feature
        """
        self._features.add(implementation)

    def get_feature(
        self, pattern: str, filename: str, line_number: int
    ) -> Optional[Tuple[Callable[..., Any], Dict[str, Any]]]:
        """
        Get the implementation for a @feature.

        Args:
            pattern: the current pattern of a feature name
            filename: the filename of the current feature
            line_number: the line number of the current feature

        Returns:
            None or tuple of (function, parameters)
        """
        matches: List[Tuple[Callable[..., Any], Dict[str, Any]]] = []

        for func in self._features:
            # try to match and convert the parameters
            for p in func._htf_features:  # type: ignore
                match = p.match(pattern)
                if match is not None:
                    matches.append((func, match))

        if len(matches) == 0:
            return None

        if len(matches) >= 2:
            raise Exception(
                "Multiple @feature implementations found for pattern '{}' in '{}:{}'".format(
                    pattern, filename, line_number
                )
                + "\n\n"
                + self.get_matches_printable(matches)
            )

        # we found one match
        return matches[0]

    def register_scenario(self, implementation: Callable[..., Any]) -> None:
        """
        Register a @scenario.

        Args:
            implementation: implementation of a @scenario
        """
        self._scenarios.add(implementation)

    def get_scenario(
        self, pattern: str, filename: str, line_number: int
    ) -> Optional[Tuple[Callable[..., Any], Dict[str, Any]]]:
        """
        Get the implementation for a @scenario.

        Args:
            pattern: the current pattern of a scenario name
            filename: the filename of the current feature
            line_number: the line number of the current scenario

        Returns:
            None or tuple of (function, parameters)
        """
        matches = []

        for func in self._scenarios:
            # try to match and convert the parameters
            for p in func._htf_scenarios:  # type: ignore
                match = p.match(pattern)
                if match is not None:
                    matches.append((func, match))

        if len(matches) == 0:
            return None

        if len(matches) >= 2:
            raise Exception(
                "Multiple @scenario implementations found for pattern '{}' in '{}:{}'".format(
                    pattern, filename, line_number
                )
            )

        # we found one match
        return matches[0]

    def register_given_precondition(self, precondition: Callable[..., Any]) -> None:
        """
        Register a @given precondition.

        Args:
            precondition: implementation of a @given step
        """
        self._given_preconditions.add(precondition)

    def get_given_precondition(
        self, precondition: str, filename: str, line_number: int
    ) -> Optional[Tuple[Callable[..., Any], Dict[str, Any]]]:
        """
        Get the implementation for a @given precondition.

        Args:
            precondition: the current precondition from a feature file.
            filename: the filename of the current feature
            line_number: the line number of the current precondition

        Returns:
            tuple of (function, parameters)
        """
        matches = []

        for func in self._given_preconditions:
            # try to match and convert the parameters
            for p in func._htf_given_preconditions:  # type: ignore
                match = p.match(precondition)
                if match is not None:
                    matches.append((func, match))

        # assert that exactly one match was found
        if len(matches) == 0:
            raise Exception(
                "No @given implementation found for precondition '{}' in '{}:{}'".format(
                    precondition, filename, line_number
                )
            )

        if len(matches) >= 2:
            raise Exception(
                "Multiple @given implementations found for precondition '{}' in '{}:{}'".format(
                    precondition, filename, line_number
                )
                + "\n\n"
                + self.get_matches_printable(matches)
            )

        # we found one match
        return matches[0]

    def register_when_condition(self, condition: Callable[..., Any]) -> None:
        """
        Register a @when condition.

        Args:
            condition: implementation of a @when step
        """
        self._when_conditions.add(condition)

    def get_when_condition(
        self, condition: str, filename: str, line_number: int
    ) -> Optional[Tuple[Callable[..., Any], Dict[str, Any]]]:
        """
        Get the implementation for a @when condition.

        Args:
            condition: the current condition from a feature file.
            filename: the filename of the current feature
            line_number: the line number of the current condition

        Returns:
            tuple of (function, parameters)
        """
        matches = []

        for func in self._when_conditions:
            # try to match and convert the parameters
            for p in func._htf_when_conditions:  # type: ignore
                match = p.match(condition)
                if match is not None:
                    matches.append((func, match))

        # assert that exactly one match was found
        if len(matches) == 0:
            raise Exception(
                "No @when implementation found for condition '{}' in '{}:{}'".format(condition, filename, line_number)
            )

        if len(matches) >= 2:
            raise Exception(
                "Multiple @when implementations found for condition '{}' in '{}:{}'".format(
                    condition, filename, line_number
                )
                + "\n\n"
                + self.get_matches_printable(matches)
            )

        # we found one match
        return matches[0]

    def register_then_assertion(self, assertion: Callable[..., Any]) -> None:
        """
        Register a @then condition.

        Args:
            assertion: implementation of a @given step
        """
        self._then_assertions.add(assertion)

    def get_then_condition(
        self, assertion: str, filename: str, line_number: int
    ) -> Optional[Tuple[Callable[..., Any], Dict[str, Any]]]:
        """
        Get the implementation for a @then assertion.

        Args:
            assertion: the current assertion from a feature file.
            filename: the filename of the current feature
            line_number: the line number of the current assertion
        """
        matches = []

        for func in self._then_assertions:
            # try to match and convert the parameters
            for p in func._htf_then_assertions:  # type: ignore
                match = p.match(assertion)
                if match is not None:
                    matches.append((func, match))

        # assert that exactly one match was found
        if len(matches) == 0:
            raise Exception(
                "No @then implementation found for assertion '{}' in '{}:{}'".format(assertion, filename, line_number)
            )

        if len(matches) >= 2:
            raise Exception(
                "Multiple @then implementations found for assertion '{}' in '{}:{}'".format(
                    assertion, filename, line_number
                )
                + "\n\n"
                + self.get_matches_printable(matches)
            )

        # we found one match
        return matches[0]

    def register_transformation(self, transformer: Callable[..., Any]) -> None:
        """
        Register a @transformation data transformation handler.

        Args:
            transformer: implementation of a @transformations data transformation handler
        """
        self._transformations.add(transformer)

    def get_transformation(self, name: str, filename: str, line_number: int) -> Optional[Callable[..., Any]]:
        """
        Get the implementation for a @transformation data transformation handler.

        Args:
            name: the current name of the example from a feature file.
            filename: the filename of the current feature
            line_number: the line number of the current assertion
        """
        matches = []

        for func in self._transformations:
            # try to match and convert the parameters
            for p in func._htf_transformations:  # type: ignore
                match = p.match(name)
                if match is not None:
                    matches.append((func, match))

        if len(matches) == 0:
            return None

        if len(matches) >= 2:
            raise Exception(
                "Multiple @transformation implementations found for pattern '{}' in '{}:{}'".format(
                    name, filename, line_number
                )
                + "\n\n"
                + self.get_matches_printable(matches)
            )

        # we found one match
        return matches[0][0]  # no parameters needed

    def register_data_generator(self, data_generator: Callable[..., Any]) -> None:
        """
        Register a @data_generator data generator.

        Args:
            data_generator: implementation of a @data_generator data generator
        """
        self._data_generators.add(data_generator)

    def get_data_generator(self, title: str, filename: str, line_number: int) -> Optional[Callable[..., Any]]:
        """
        Get the implementation for a @data_generator data generator.

        Args:
            title: the current title of the generator from a feature file.
            filename: the filename of the current feature
            line_number: the line number of the current assertion
        """
        matches = []

        for func in self._data_generators:
            # try to match and convert the parameters
            for p in func._htf_data_generators:  # type: ignore
                match = p.match(title)
                if match is not None:  # pragma: no cover
                    matches.append((func, match))

        # assert that exactly one match was found
        if len(matches) == 0:
            raise Exception(
                "No @data_generator implementation found for title '{}' in '{}:{}'".format(title, filename, line_number)
            )

        if len(matches) >= 2:
            raise Exception(
                "Multiple @data_generator implementations found for pattern '{}' in '{}:{}'".format(
                    title, filename, line_number
                )
                + "\n\n"
                + self.get_matches_printable(matches)
            )

        # we found one match
        return matches[0][0]  # no parameters needed

    def combine(self, other: "GherkinSteps") -> None:
        self._features |= other._features
        self._scenarios |= other._scenarios
        self._given_preconditions |= other._given_preconditions
        self._when_conditions |= other._when_conditions
        self._then_assertions |= other._then_assertions
        self._transformations |= other._transformations
        self._data_generators |= other._data_generators
