#
# Copyright (c) 2024, HILSTER - https://hilster.io
# All rights reserved.
#

from typing import Iterable, AsyncGenerator, Optional, Any, Generator, List, Callable, Union
import asyncio

from htf.gherkin.gherkin_steps import GherkinSteps
from htf.gherkin.scenario import Scenario
from htf.gherkin.step import Step


async def iterate(iterable: Iterable) -> AsyncGenerator[Any, None]:
    """
    Iterate sync and async iterables
    """
    if hasattr(iterable, "__aiter__"):
        # async iterator
        async for i in iterable:  # pragma: no branch
            yield i
    else:
        # sync iterator
        for i in iterable:
            yield i


class DataGenerator:
    def __init__(
        self,
        generator: Iterable,
        filename: str,
        line_number: Optional[int] = None,
    ) -> None:
        self.generator = generator
        self.filename = filename
        self.line_number = line_number


class PythonDataGenerator:
    def __init__(
        self,
        title: str,
        filename: str,
        line_number: Optional[int] = None,
    ) -> None:
        self.title = title
        self.generator: Optional[Generator] = None
        self.filename = filename
        self.line_number = line_number

    def set_generator(self, generator: Callable[..., Generator]) -> None:
        """
        Search for the data generator and setup generator.
        """
        self.generator = generator()


class ScenarioOutline:
    def __init__(
        self,
        name: str,
        doc: str,
        filename: str,
        line_number: int,
    ) -> None:
        self.name = name
        self.doc = doc
        self.examples_name: Optional[str] = None
        self.transformation = None
        self.filename = filename
        self.line_number = line_number

        self.given_preconditions: List[Step] = list()
        self.when_conditions: List[Step] = list()
        self.then_assertions: List[Step] = list()

        self.data_generators: List[Union[DataGenerator, PythonDataGenerator]] = []

        # Tags, requirements and skip reasons
        self.tags: Optional[List[str]] = None
        self.requirements: Optional[List[str]] = None
        self.skip_reason: Optional[str] = None

    def set_examples_name(self, name: str) -> None:
        self.examples_name = name

    def add_given(self, step: Step) -> None:
        self.given_preconditions.append(step)

    def add_when(self, step: Step) -> None:
        self.when_conditions.append(step)

    def add_then(self, step: Step) -> None:
        self.then_assertions.append(step)

    def add_data_generator(self, data_generator: Union[DataGenerator, PythonDataGenerator]) -> None:
        assert data_generator is not None
        self.data_generators.append(data_generator)

    def setup_transformation(self, gherkin_steps: GherkinSteps) -> None:
        self.transformation = gherkin_steps.get_transformation(  # type: ignore
            self.examples_name, self.filename, self.line_number  # type: ignore
        )  # no parameters needed here

    def setup_data_generators(self, gherkin_steps: GherkinSteps) -> None:
        for generator in self.data_generators:
            if isinstance(generator, PythonDataGenerator):
                # the generator needs setup
                generator.set_generator(
                    gherkin_steps.get_data_generator(
                        generator.title, generator.filename, generator.line_number  # type: ignore
                    )
                )

    async def __aiter__(self) -> AsyncGenerator[Scenario, None]:
        """
        Generate Scenarios with replacements.
        """

        for data_generator in self.data_generators:
            async for replacements in iterate(data_generator.generator):  # type: ignore  # pragma: no branch
                # transform data
                if self.transformation:
                    if asyncio.iscoroutinefunction(self.transformation):
                        replacements = await self.transformation(replacements)
                    else:
                        replacements = self.transformation(replacements)

                scenario = Scenario(
                    name=self.name, doc="", filename=self.filename, line_number=self.line_number, kwargs=replacements
                )

                # set tags and requirements
                scenario.set_tags(self.tags)  # type: ignore
                scenario.set_requirements(self.requirements)  # type: ignore
                scenario.set_skipped(self.skip_reason)  # type: ignore

                # deep copy data and replace everything
                for given_precondition in self.given_preconditions:
                    scenario.add_given(given_precondition.replace(replacements, data_generator.filename))
                for when_condition in self.when_conditions:
                    scenario.add_when(when_condition.replace(replacements, data_generator.filename))
                for then_assertion in self.then_assertions:
                    scenario.add_then(then_assertion.replace(replacements, data_generator.filename))

                yield scenario

    def set_tags(self, tags: List[str]) -> None:
        self.tags = tags

    def set_requirements(self, requirements: List[str]) -> None:
        self.requirements = requirements

    def set_skipped(self, reason: str) -> None:
        self.skip_reason = reason
