#
# Copyright (c) 2024, HILSTER - https://hilster.io
# All rights reserved.
#

from typing import Optional, Generator
import socket

from werkzeug.serving import make_server  # type: ignore
import threading
import webbrowser
import pyqrcode  # type: ignore
import platform
from htf.core import BrowserInteraction
from htf.termcolors import info


class BrowserInteractionServer(threading.Thread):
    def __init__(
        self,
        app: BrowserInteraction,
        host: str = "0.0.0.0",
        port: int = 8080,
        url: str = "/",
        open_browser: bool = True,
    ) -> None:
        threading.Thread.__init__(self)
        self.server = make_server(host, port, app, threaded=True)
        self.context = app.app_context()
        self.context.push()
        hostname = socket.gethostname()
        full_url = "http://{}:{}{}".format(hostname, port, url)

        if full_url.lower().endswith("/index.html"):
            full_url = full_url[:-11]

        if full_url.lower().endswith("/"):
            full_url = full_url[:-1]

        print(info("Running an interactive session on {}".format(full_url)))
        qrcode = pyqrcode.create(full_url)

        if platform.system() == "Windows":
            # Windows
            print(qrcode.terminal(module_color="black", background="white", quiet_zone=1))
        else:
            # Linux, etc.
            print(qrcode.terminal(quiet_zone=1))

        if open_browser is True:
            webbrowser.open(full_url)

    def run(self) -> None:
        self.server.serve_forever()

    def shutdown(self) -> None:
        self.server.shutdown()


def BrowserInteractionFixture(  # noqa: N802
    data_directory: Optional[str] = None,
    host: str = "0.0.0.0",
    port: int = 8080,
    url: str = "/",
    open_browser: bool = True,
) -> Generator[BrowserInteraction, None, None]:
    interaction = BrowserInteraction(__name__)

    if data_directory is not None:
        interaction.set_data_directory(data_directory)

    server = BrowserInteractionServer(interaction, host=host, port=port, url=url, open_browser=open_browser)
    server.start()
    yield interaction
    server.shutdown()
