#
# Copyright (c) 2024, HILSTER - https://hilster.io
# All rights reserved.
#

from typing import Dict, List

import glob
import os
import json
import subprocess
import sys
from htf.termcolors import info, system
from htf.core import get_packages_from_pip_freeze_txt, get_installed_packages, get_data_file


def merge_licenses(a: List[Dict[str, str]], b: List[Dict[str, str]]) -> List[Dict[str, str]]:
    """
    Merge licenses into one.
    """
    result = list(sorted(a + b, key=lambda x: x["Name"]))
    return result


def unique_licenses(products_and_licenses: Dict[str, List[Dict[str, str]]]) -> Dict[str, List[Dict[str, str]]]:
    """
    Make licenses unique.
    """

    unique_licenses: Dict[str, Dict[str, str]] = {}

    for name, licenses in products_and_licenses.items():
        for license_ in licenses:
            name_and_version = license_["Name"] + " " + license_["Version"]
            if name_and_version not in unique_licenses:
                unique_licenses[name_and_version] = license_

    return {"all": list(sorted(unique_licenses.values(), key=lambda k: (k["Name"], k["Version"])))}


def get_licenses_from_file(filename: str) -> List[Dict[str, str]]:
    """
    Load a licenses file.
    """
    with open(filename, "r", encoding="utf-8") as fp:
        licenses = json.load(fp)

    for license_ in licenses:
        license_["LicenseText"] = license_["LicenseText"].strip().strip("\n")
        while license_["LicenseText"].endswith("\n"):  # pragma: no branch
            license_["LicenseText"] = license_["LicenseText"][:-1]  # pragma: no cover

    return licenses  # type: ignore


def get_pip_licenses_output() -> List[Dict[str, str]]:
    output = subprocess.check_output(
        "pip-licenses --format json --with-license-file --no-license-path --with-authors --with-maintainers "
        "--with-urls --with-description",
        shell=True,
    )

    data = json.loads(output)

    return data  # type: ignore


def get_filtered_pip_licenses_output(included_packages: List[str]) -> List[Dict[str, str]]:
    """
    Return all installed packages that are included in ``included_packages``.
    """
    return [license_ for license_ in get_pip_licenses_output() if license_["Name"] in included_packages]


def get_licenses(*filenames: str) -> List[Dict[str, str]]:
    """
    Read and merge licenses from *filenames and return them sorted.
    """
    licenses: List[Dict[str, str]] = []

    for filename in filenames:
        licenses = merge_licenses(licenses, get_licenses_from_file(filename))

    return licenses


def print_product_licenses(product_name: str, licenses: List[Dict[str, str]], compact: bool) -> None:
    """
    Print out all licenses for a product.
    """
    if not compact:
        tmp = f"Licenses for {product_name}"
    else:
        tmp = "Licenses"

    print(system(tmp + "\n" + "=" * len(tmp)) + "\n")

    for license_ in licenses:
        name_and_version = license_["Name"] + " " + license_["Version"]
        if compact:
            print(info(name_and_version))
        else:
            print(info(name_and_version + "\n" + "-" * len(name_and_version) + "\n"))
            for key in ["Description", "Author", "Maintainer", "URL", "License", "LicenseText"]:
                newline = "\n" if key == "LicenseText" else " "
                left_part = info(f"{key + ':':16}")

                left_part = "" if key == "LicenseText" else left_part

                print(f"{left_part}{newline}{license_[key]}")

            print()

    if compact:
        print()


def indent(text: str, indentation: int = 4) -> str:
    lines = text.split("\n")

    return " " * indentation + ("\n" + " " * indentation).join(lines)


def print_product_licenses_markdown(product_name: str, licenses: List[Dict[str, str]], compact: bool) -> None:
    """
    Print licenses in Markdown format used as an SWBOM.
    """
    if not compact:
        print(f"# Licenses for {product_name}\n")
    else:
        print("# Licenses\n")

    for license_ in licenses:
        if compact:
            print("- " + license_["Name"] + " " + license_["Version"])
        else:
            print("## " + license_["Name"] + " " + license_["Version"] + "\n")

            for key in ["Description", "Author", "Maintainer", "URL", "License", "LicenseText"]:
                newline = "\n" if key == "LicenseText" else " "
                left_part = "**" + key + ":**"

                left_part = "" if key == "LicenseText" else left_part

                if key == "LicenseText":
                    license_[key] = indent(license_[key], 4)

                print(f"{left_part}{newline}{license_[key]}\n")

            print()

    if compact:
        print()


def get_product_requirements(name: str) -> List[Dict[str, str]]:
    """
    Return a list of packages shipped with requirements.txt for a given product.
    """

    requirements_txt = get_data_file(
        os.path.join(name, "requirements.txt"),
        os.path.join(os.path.dirname(__file__), "..", "requirements.txt"),  # for development only
    )

    if requirements_txt is None:  # pragma: no branch
        return []  # pragma: no cover

    required_packages = get_packages_from_pip_freeze_txt(requirements_txt)

    installed_packages = get_installed_packages()

    installed_required_packages = []

    for package_name in required_packages:
        if package_name in installed_packages:  # pragma: no branch
            installed_required_packages.append(package_name)

    installed_required_packages = list(sorted(installed_required_packages))

    return get_filtered_pip_licenses_output(installed_required_packages)


def collect_licenses(compact: bool) -> Dict[str, List[Dict[str, str]]]:
    """
    Collect all product licenses.
    """
    # collect currently used licenses using pip-licenses and filter by sys.prefix + /htf/requirements.txt
    # iterate over sys.prefix/build-information/**/*.json
    # group products by **
    # add currently installed packages found in sys.prefix/<product>/requirements.txt
    # join all licenses
    # sort by product.Name
    licenses: Dict[str, List[Dict[str, str]]] = {}

    for product in sorted(os.listdir(os.path.join(sys.prefix, "build-information"))):
        license_files = glob.glob(os.path.join(sys.prefix, "build-information", product, "*.json"))

        licenses[product] = merge_licenses(
            get_licenses(*license_files),
            get_product_requirements(product),
        )

    if compact:
        licenses = unique_licenses(licenses)

    return licenses


def print_licenses(compact: bool) -> None:
    """
    Print all licenses to console.
    """
    products_and_licenses: Dict[str, List[Dict[str, str]]] = collect_licenses(compact)

    # print licenses starting with htf
    if "htf" in products_and_licenses:
        print_product_licenses("htf", products_and_licenses["htf"], compact)

    # print all other product licenses in order
    for name, licenses in products_and_licenses.items():
        if name == "htf":
            continue
        print_product_licenses(name, licenses, compact)


def print_sbom(compact: bool) -> None:
    """
    Create SBOM in Markdown format and print it.
    """
    products_and_licenses: Dict[str, List[Dict[str, str]]] = collect_licenses(compact)

    # print licenses starting with htf
    if "htf" in products_and_licenses:  # pragma: no branch
        print_product_licenses_markdown("htf", products_and_licenses["htf"], compact)

    # print all other product licenses in order
    for name, licenses in products_and_licenses.items():
        if name == "htf":
            continue
        print_product_licenses_markdown(name, licenses, compact)
