#
# Copyright (c) 2024, HILSTER - https://hilster.io
# All rights reserved.
#

from typing import Any, Callable


class DelegationError(Exception):
    """
    ``DelegationError`` is thrown if :class:`DelegateMixin` is used directly without being mixed in.
    """


class DelegateMixin:
    """
    ``DelegateMixin`` defines :func:`set` and :func:`get` as methods that have
    to be overwritten by subclass.
    """

    def set(self, value: Any) -> None:
        """
        The setter for a delegate.

        Args:
            value: the value to be set.

        Note:
            This method is abstract.
        """
        raise DelegationError("set() should never be called! You have to overwrite it!")

    def get(self) -> Any:
        """
        The getter for a delegate.

        Note:
            This method is abstract.
        """
        raise DelegationError("get() should never be called! You have to overwrite it!")


class DelegateCallbackMixin:
    """
    ``DelegateCallbackMixin`` gives the possibility to user define
    setter and getter methods using a :class:`DelegateMixin`.
    """

    def __init__(self, set_callback: Callable, get_callback: Callable) -> None:
        """
        Args:
            set_callback: the method to be called as a setter.
            get_callback: the method to be called as a getter.
        """
        super(DelegateCallbackMixin, self).__init__()

        self.set = set_callback
        self.get = get_callback


class DelegatorMixin:
    """
    The ``DelegatorMixin`` automatically delegates calls for :func:`DelegateMixin.set` and
     :func:`DelegateMixin.get` to its members if they are instances of :class:`DelegateMixin`.
    """

    def __getattribute__(self, name: str) -> Any:
        try:
            delegate = object.__getattribute__(self, name)
            if not isinstance(delegate, (DelegateMixin, DelegateCallbackMixin)):
                raise AttributeError("Can only delegate to instances of Delegate!")
            return delegate.get()
        except AttributeError:
            pass
        return object.__getattribute__(self, name)

    def __setattr__(self, name: str, value: Any) -> Any:
        try:
            delegate = object.__getattribute__(self, name)
            if not isinstance(delegate, (DelegateMixin, DelegateCallbackMixin)):
                raise AttributeError("Can only delegate to instances of Delegate!")
            return delegate.set(value)
        except AttributeError:
            pass

        object.__setattr__(self, name, value)
