#
# Copyright (c) 2024, HILSTER - https://hilster.io
# All rights reserved.
#

from typing import Any, Dict
import os
import json
import copy

import requests
from htf.termcolors import system


class JSONUnicodeEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        if not isinstance(obj, bytes):
            return obj.encode("utf-8", errors="ignore")

        return json.JSONEncoder.default(self, obj)


class ReportServerReport:
    """
    ``ReportServerReport`` report a test report to a report server.
    """

    def __init__(self, url: str) -> None:
        """
        Initialize a ``JSONTestReport`` instance.

        Args:
            filename: the filename for the json test report.
        """
        self.url = url
        self.report_type = "REPORT SERVER"

    def render(self, data: Dict[str, Any]) -> None:
        """
        Generates and sends a test report.

        Args:
            data: test result data
        """
        data = copy.deepcopy(data)

        # convert tags into a list
        for test in data["tests"]:
            try:
                test["data"]["tags"] = list(sorted(test["data"]["tags"]))
            except KeyError:
                pass

        # escape all data items
        def safe_escape(data: Any) -> Any:
            if isinstance(data, dict):
                ret = {k: safe_escape(v) for k, v in list(data.items())}
                data.update(ret)
                return data
            elif isinstance(data, list):
                return [safe_escape(item) for item in data]
            else:
                if isinstance(data, bytes):
                    return data.encode("utf-8", errors="ignore")  # type: ignore
                return data

        data = safe_escape(data)

        self.write({"origin": "htf", "report": data})

    def write(self, data: Dict[str, Any]) -> None:
        verify = True

        if "HTF_INSECURE" in os.environ:
            verify = False
        try:
            response = requests.post(self.url, json=data, verify=verify)
            assert response.status_code == 204, "The report server {} returned {}".format(
                self.url, response.status_code
            )
        except Exception as e:
            print(system("An error occurred when sending sending the test report to {}: {}".format(self.url, e)))
        else:
            print(system("Sent report to {}".format(self.url)))
