#
# Copyright (c) 2024, HILSTER - https://hilster.io
# All rights reserved.
#

from typing import Optional, Dict, Any, Generator, Tuple, List
import datetime
import platform
import sys
import time
import uuid

import hlm

import htf
from htf.result import TestResult

from htf.event import EventManager, Event
from htf.exceptions import RepeatTest
from htf.runner import TestRunner


class ReportData:
    def __init__(
        self, title: str, metadata: Optional[Dict[str, Any]] = None, minimize_report_size: bool = False
    ) -> None:
        EventManager.register_callback("session_started", self.handle_session_started)
        EventManager.register_callback("result_finished", self.handle_result_finished)
        EventManager.register_callback("session_finished", self.handle_session_finished)
        self._start_time = None
        self._stop_time = None
        self._overwrite_results = False
        self._test_number = 0
        self._minimize_report_size = minimize_report_size

        self.metadata = metadata or {}
        self.title = title
        self._results: List[TestResult] = []
        self._data = {
            "id": str(uuid.uuid4()),
            "title": self.title,
            "unexpected_successes": 0,
            "expected_failures": 0,
            "metadata": self.metadata,
            "tests": [],
            "status": "draft",
            "htf_version": htf.__version__,
            "python_version": platform.python_version(),
            "license_information": hlm.get_used_licenses_formatted(),
            "platform": platform.platform(),
            "hostname": platform.node(),
            "machine": platform.machine(),
            "test_amount": 0,
            "successes": 0,
            "errors": 0,
            "failures": 0,
            "skipped": 0,
        }

    @staticmethod
    def _get_htf_command() -> str:
        # add htf call for reproduction
        def as_command(part: str) -> str:
            part = str(part)
            if " " in part:
                return '"{}"'.format(part)
            else:
                return part

        return " ".join(map(as_command, sys.argv))

    @staticmethod
    def _get_system_verification_report() -> Optional[str]:
        import htf.core

        return htf.core.get_system_verification_report()

    async def schedule(self, runner: TestRunner) -> None:
        await runner.run()
        self._overwrite_results = True

    def handle_session_started(self, event: Event) -> None:
        self._start_time = event.kwargs["time"]

    def update_results_data(self) -> None:
        """
        Update test data after a result is finished or all tests are finished.
        """
        self.metadata["HTF Command"] = self._get_htf_command()

        system_verification_report = self._get_system_verification_report()
        if system_verification_report is not None:  # pragma: no branch
            self.metadata["System Verification Report"] = system_verification_report  # pragma: no cover (manual test)

        self._data.update({"date": datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"), "metadata": self.metadata})

        if self._start_time is not None:
            self._data.update(
                {
                    "start_time": self._start_time.isoformat(),
                    "start_timestamp": int(time.mktime(self._start_time.timetuple())),
                }
            )

        if self._stop_time is not None:
            self._data.update(
                {
                    "stop_time": self._stop_time.isoformat(),
                    "stop_timestamp": int(time.mktime(self._stop_time.timetuple())),
                }
            )

        if self._start_time is not None and self._stop_time is not None:
            self._data.update(
                {
                    "duration": str(self._stop_time - self._start_time).split(".")[0],
                    "seconds": (self._stop_time - self._start_time).total_seconds(),
                }
            )
        else:
            self._data.update({"duration": "-", "seconds": "-"})

    def handle_result_finished(self, event: Event) -> None:
        if self._overwrite_results:  # actual test results
            result_dict = event.kwargs["result_object"].to_dict()
            if self._minimize_report_size:
                if result_dict["result"] in ["success", "skipped"]:
                    if result_dict["data"]["stdout"]:  # pragma: no branch
                        result_dict["data"]["stdout"] = "(stdout was removed for size minimization)"
                    if result_dict["data"]["stderr"]:
                        result_dict["data"]["stderr"] = "(stderr was removed for size minimization)"
            if event.kwargs["result_object"].exception_type is RepeatTest:
                self._results.insert(self._test_number, event.kwargs["result_object"])
                self._data["tests"].insert(self._test_number, result_dict)  # type: ignore
            else:
                self._results[self._test_number] = event.kwargs["result_object"]
                self._data["tests"][self._test_number] = result_dict  # type: ignore
            self._test_number += 1
        else:  # add scheduled results
            self._results.append(event.kwargs["result_object"])
            self._data["tests"].append(event.kwargs["result_object"].to_dict())  # type: ignore

        self.update_results_data()

        self._data["test_amount"] = len(self._results)
        self._data["successes"] = len([res for res in self._results if res.result == "success"])
        self._data["errors"] = len([res for res in self._results if res.result == "error"])
        self._data["failures"] = len([res for res in self._results if res.result == "failure"])
        self._data["skipped"] = len([res for res in self._results if res.result == "skipped"])

    def handle_session_finished(self, event: Event) -> None:
        self._stop_time = event.kwargs["time"]
        self.update_results_data()
        self._data.update({"status": "final"})

    def __iter__(self) -> Generator[Tuple[str, Any], None, None]:
        for key, value in self._data.items():
            yield key, value
