#
# Copyright (c) 2024, HILSTER - https://hilster.io
# All rights reserved.
#

from typing import (
    Optional,
    Generator,
    Callable,
    Any,
    Tuple,
    Dict,
    Union,
    AsyncGenerator,
    ContextManager,
    AsyncContextManager,
)
import asyncio

import time
from functools import wraps
from contextlib import contextmanager, asynccontextmanager


def attempts_generator(attempts: int = 2, delay: Optional[float] = None) -> Generator[int, None, None]:
    """
    A generator for attempts with a configurable delay.

    Args:
        attempts=2: the number of attempts
        delay=None: the delay between attempts in seconds

    Returns:
        a generator that generates ``attempts`` attempts with the given delay between them
    """

    delay = delay or 0

    for attempt in range(attempts):
        yield attempt

        if attempt < attempts - 1:
            time.sleep(delay)


async def retry(
    action: Callable[..., Any],
    args: Optional[Tuple] = None,
    kwargs: Optional[Dict[str, Any]] = None,
    attempts: int = 2,
    delay: Optional[float] = None,
    retry_exceptions: Tuple[Any] = (BaseException,),
) -> Any:
    """
    Retry ``action`` with a given number of attempts and a given delay.

    Args:
        action: the action to be retried (coroutines are supported)
        args=(): the positional arguments passed to ``action``
        kwargs={}: the keyword arguments passed to ``action``
        attempts=2: the number of attempts
        delay=None: the delay between attempts in seconds
        retry_exceptions=(BaseException,): the exceptions to be catched to retry the action

    Returns:
        the return value of ``action(*args, **kwargs)``
    """

    assert callable(action)

    if args is None:
        args = ()

    if kwargs is None:
        kwargs = {}

    n = 1

    for _ in attempts_generator(attempts=attempts, delay=delay):
        try:
            if asyncio.iscoroutinefunction(action):
                return await action(*args, **kwargs)
            else:
                return action(*args, **kwargs)
        except retry_exceptions:
            if n == attempts:
                raise
        finally:
            n += 1


def retry_sync(
    action: Callable[..., Any],
    args: Optional[Tuple] = None,
    kwargs: Optional[Dict[str, Any]] = None,
    attempts: int = 2,
    delay: Optional[float] = None,
    retry_exceptions: Tuple[Any] = (BaseException,),
) -> Any:
    """
    Retry ``action`` with a given number of attempts and a given delay.

    Args:
        action: the action to be retried (coroutines are not supported)
        args=(): the positional arguments passed to ``action``
        kwargs={}: the keyword arguments passed to ``action``
        attempts=2: the number of attempts
        delay=None: the delay between attempts in seconds
        retry_exceptions=(BaseException,): the exceptions to be catched to retry the action

    Returns:
        the return value of ``action(*args, **kwargs)``
    """

    assert callable(action)

    if args is None:
        args = ()

    if kwargs is None:
        kwargs = {}

    n = 1

    if asyncio.iscoroutinefunction(action):
        raise TypeError("Coroutines are not supported. Use htf.retry() instead!")

    for _ in attempts_generator(attempts=attempts, delay=delay):
        try:
            return action(*args, **kwargs)
        except retry_exceptions:
            if n == attempts:
                raise
        finally:
            n += 1


def retriable(
    attempts: int = 2,
    delay: Optional[float] = None,
    retry_exceptions: Tuple[Any] = (BaseException,),
) -> Callable[[Callable[..., Any]], Any]:
    """
    A decorator factory for retry(). Wrap your function in @retriable(...) to
    give it retry powers!

    Args:
        attempts=2: the number of attempts
        delay=None: the delay between attempts in seconds
        retry_exceptions=(BaseException,): the exceptions to be catched to retry the action

    Returns:
        a function decorator supporting retries. If a coroutine is supplied a coroutine is returned
            else not.
    """

    def _retriable_factory(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        async def _async_retriable_wrapper(*args: Any, **kwargs: Any) -> Any:
            return await retry(
                action=func, args=args, kwargs=kwargs, attempts=attempts, delay=delay, retry_exceptions=retry_exceptions
            )

        @wraps(func)
        def _retriable_wrapper(*args: Any, **kwargs: Any) -> Any:
            return retry_sync(
                action=func, args=args, kwargs=kwargs, attempts=attempts, delay=delay, retry_exceptions=retry_exceptions
            )

        if asyncio.iscoroutinefunction(func):
            return _async_retriable_wrapper
        else:
            return _retriable_wrapper

    return _retriable_factory


def retrying(
    action: Callable[..., Any],
    attempts: int = 2,
    delay: Optional[float] = None,
    retry_exceptions: Tuple[Any] = (BaseException,),
) -> Union[ContextManager, AsyncContextManager]:
    """
    A context manager for wrapping ``action`` with retry functionality
    with a given number of attempts and a given delay.

    Args:
        action: the action to be retried
        attempts=2: the number of attempts
        delay=None: the delay between attempts in seconds
        retry_exceptions=(BaseException,): the exceptions to be catched to retry the action

    Returns:
        a context manager supporting retries. If a coroutine is supplied an async context manager
            is returned.
    """

    @asynccontextmanager
    async def _async_retrying() -> AsyncGenerator[Callable[..., Any], None]:
        yield retriable(attempts=attempts, delay=delay, retry_exceptions=retry_exceptions)(action)

    @contextmanager
    def _retrying() -> Generator[Callable[..., Any], None, None]:
        yield retriable(attempts=attempts, delay=delay, retry_exceptions=retry_exceptions)(action)

    if asyncio.iscoroutinefunction(action):
        return _async_retrying()
    else:
        return _retrying()
