#
# Copyright (c) 2024, HILSTER - https://hilster.io
# All rights reserved.
#

from typing import List, Any, Coroutine, Callable
import asyncio
from functools import wraps

from htf.core import TestRunner as _TestRunner, TestResult, Test
from htf.core import DryRunner as _DryRunner
from htf.environment import TestEnvironment
from htf.evaluation import Evaluation
from htf.event import EventManager, Event
from htf.exceptions import raise_coroutine_runtime_warnings_as_runtime_errors


class TestRunner(_TestRunner):
    def __init__(
        self,
        test_suite: "TestSuite",  # type: ignore  # noqa: F821
        environment: TestEnvironment,
        evaluation: Evaluation,
    ) -> None:
        """
        Initialize self.  See help(type(self)) for accurate signature.
        """
        super(TestRunner, self).__init__(test_suite=test_suite, environment=environment, evaluation=evaluation)

    async def run(self) -> List[TestResult]:
        return await super(TestRunner, self).run()

    def _wrap_test(
        self, instance: "TestCase", test: Test  # type: ignore  # noqa: F821
    ) -> Callable[..., Coroutine[Any, Any, None]]:
        """
        Wrapper for legacy tests inheriting from htf.TestCase.
        Necessary to ensure set_up(), after_test() and tear_down() are called.

        Args:
            instance: test case instance
            test: test method

        Returns:
            wrapped test function
        """

        @wraps(test)
        async def wrapper(*args: Any, **kwargs: Any) -> None:
            try:
                if asyncio.iscoroutinefunction(instance.set_up):
                    await instance.set_up()
                else:
                    instance.set_up()

                if asyncio.iscoroutinefunction(test):
                    await test(*args, **kwargs)
                else:
                    test(*args, **kwargs)  # type: ignore
            finally:
                if asyncio.iscoroutinefunction(instance.tear_down):
                    await instance.tear_down()
                else:
                    instance.tear_down()

        return wrapper

    def enable_fail_fast(self) -> None:
        """
        Enables the fail_fast option by registering an event callback on 'result_finished' events.
        When a result is either 'error' or 'failure' the test run is aborted.
        """

        async def abort_on_failure(event: Event) -> None:
            if event.kwargs["result"] in ["error", "failure"]:
                self._aborted = True
                await self._tear_down()  # this may result in an extra 'class_finished'
                # event if the class was done anyway
                with raise_coroutine_runtime_warnings_as_runtime_errors():
                    await self._handle_event(Event("module_finished"))  # again, this event might be the second
                    # 'module_finished'
                # doubled events here should be acceptable, since fixtures will only be torn down once

        EventManager.register_callback("result_finished", abort_on_failure)


class DryRunner(_DryRunner):
    def __init__(
        self,
        test_suite: "TestSuite",  # type: ignore  # noqa: F821
        environment: TestEnvironment,
        evaluation: Evaluation,
    ) -> None:
        """
        Initialize self.  See help(type(self)) for accurate signature.
        """
        super(DryRunner, self).__init__(test_suite=test_suite, environment=environment, evaluation=evaluation)

    async def run(self) -> List[TestResult]:
        return await super(DryRunner, self).run()
