#
# Copyright (c) 2024, HILSTER - https://hilster.io
# All rights reserved.
#

from typing import Optional, Callable, Any, List, AsyncGenerator, Generator, Union, Set
import random

from htf.core import Test, TestType
from htf.discovery import Discovery


class TestSuite:
    """
    A generator for a collection of tests
    """

    def __init__(
        self,
        discovery: Discovery,
        tests_filter_func: Optional[Callable[..., bool]] = None,
        fixtures_filter_func: Optional[Callable[..., bool]] = None,
        shuffle: bool = False,
    ) -> None:
        """
        Args:
            discovery: a discovery instance
            tests_filter_func: optional filter function for tests
            fixtures_filter_func: optional filter function for fixtures
            shuffle=False: set to ``True`` to shuffle tests
        """
        self._tests = discovery.discover_tests()
        self._fixtures: List[Union[Callable[..., Any], Generator[Any, None, None], AsyncGenerator[Any, None]]] = (
            discovery.discover_fixtures()
        )
        self._tests_filter_func = tests_filter_func
        self._fixtures_filter_func = fixtures_filter_func
        self._length: Optional[int] = None
        self._shuffle = shuffle
        self._expanded_tests: Optional[List[Test]] = None

    async def __aiter__(self) -> AsyncGenerator[Test, None]:
        if self._expanded_tests is None:  # pragma: no branch
            self._expanded_tests = [
                expanded_test async for expanded_test in self._tests_generator(self._tests)  # type: ignore
            ]  # pragma: no cover  # noqa: E501
        if self._shuffle is True:
            random.shuffle(self._expanded_tests)

        for test in self._expanded_tests:
            yield test

    def set_tests_filter(self, filter_func: Callable[..., bool]) -> None:
        """
        Sets the TestSuites filter function for tests. If the filter function evaluates to False for a given test,
        the test is not yielded.

        Args:
            filter_func: filter function
        """
        assert callable(filter_func)
        self._tests_filter_func = filter_func

    def set_fixtures_filter(self, filter_func: Callable[..., bool]) -> None:
        """
        Sets the TestSuites filter function for fixtures. If the filter function evaluates to False for a given test,
        the test is not yielded.

        Args:
            filter_func: filter function
        """
        assert callable(filter_func)
        self._fixtures_filter_func = filter_func

    async def _tests_generator(self, tests: List[TestType]) -> AsyncGenerator[TestType, None]:
        """
        A generator function iterating over tests and applying a filter function if one is set.

        Args:
            tests: a list of test functions
        """
        for test in tests:
            if test.is_data_driven():
                async for data_driven_test in test.get_data_driven_tests():  # type: ignore  # pragma: no branch
                    if self._tests_filter_func is None or self._tests_filter_func(  # pragma: no branch
                        data_driven_test, data_driven_test.get_tags()
                    ):
                        yield data_driven_test
            else:
                if self._tests_filter_func is None or self._tests_filter_func(test, test.get_tags()):
                    yield test

    def get_fixtures(self) -> List[Union[Callable[..., Any], Generator[Any, None, None], AsyncGenerator[Any, None]]]:
        """
        Returns a list of all fixtures found by the Discovery.

        Returns:
            List of fixtures
        """

        if self._fixtures_filter_func is not None:
            # filter fixtures by given tags expression
            filtered_fixtures = []

            for fixture in self._fixtures:
                current_tags = self.get_tags_from_fixture(fixture)

                if self._fixtures_filter_func(fixture, current_tags):
                    filtered_fixtures.append(fixture)

            return filtered_fixtures
        else:
            # just return the collected fixtures
            return self._fixtures

    @staticmethod
    def get_tags_from_fixture(
        fixture: Union[Callable[..., Any], Generator[Any, None, None], AsyncGenerator[Any, None]]
    ) -> Set[str]:
        tags = set()

        if hasattr(fixture, "_htf_tags"):
            tags |= fixture._htf_tags

        return tags

    async def __alen__(self) -> Optional[int]:  # pragma: no cover
        """
        The length of a TestSuite is the number of tests it contains.

        Returns:
            Number of tests in the suite
        """
        if self._length is None:
            self._expanded_tests = list(await self._tests_generator(self._tests))  # type: ignore
            self._length = len(self._expanded_tests)  # type: ignore
        return self._length
