#
# Copyright (c) 2024, HILSTER - https://hilster.io
# All rights reserved.
#

from typing import Any, List, Optional

import ply.lex as lex  # type: ignore
import ply.yacc as yacc  # type: ignore
from htf.exceptions import ParsingError


class _TagsParser:
    tokens = (
        "NAME",
        "AND",
        "OR",
        "XOR",
        "NOT",
        "LPAREN",
        "RPAREN",
    )

    # Tokens
    t_AND = r"&"  # noqa: N815
    t_OR = r"\|"  # noqa: N815
    t_NOT = r"!"  # noqa: N815
    t_XOR = r"\^"  # noqa: N815
    t_LPAREN = r"\("  # noqa: N815
    t_RPAREN = r"\)"  # noqa: N815
    t_NAME = r"[a-zA-Z0-9_\-\.\=]+"  # noqa: N815

    # Ignored characters
    t_ignore = " \t\n\r"

    def t_error(self, t: Any) -> None:
        raise ParsingError("Illegal character '%s'" % t.value[0])

    # Parsing rules

    precedence = (
        ("left", "AND", "OR", "XOR"),
        ("right", "NOT"),
    )

    @staticmethod
    def p_statement_expr(t: List[Any]) -> None:
        """statement : expression"""
        t[0] = t[1]

    @staticmethod
    def p_expression_binop(t: List[Any]) -> None:
        """expression : expression AND expression
        | expression OR expression
        | expression XOR expression"""
        if t[2] == "&":
            t[0] = t[1] & t[3]
        elif t[2] == "|":
            t[0] = t[1] | t[3]
        else:  # t[2] == '^'
            t[0] = t[1] ^ t[3]

    @staticmethod
    def p_expression_not(t: List[Any]) -> None:
        """expression : NOT expression"""
        t[0] = not t[2]

    @staticmethod
    def p_expression_group(t: List[Any]) -> None:
        """expression : LPAREN expression RPAREN"""
        t[0] = t[2]

    def p_expression_name(self, t: List[Any]) -> None:
        """expression : NAME"""
        t[0] = t[1] in self.tag_list  # type: ignore

    def p_error(self, t: List[Any]) -> None:
        try:
            raise ParsingError("Syntax error at '%s'" % t.value)  # type: ignore
        except AttributeError:
            raise ParsingError("Syntax error in '%s'" % self._expression)

    def __init__(self, expression: str) -> None:
        self.tag_list: Optional[List[str]] = None
        self.lexer = lex.lex(module=self)
        self.parser = yacc.yacc(module=self, debug=False, write_tables=False)
        self._expression = expression

    def match(self, tags: List[str]) -> bool:
        self.tag_list = tags
        result = self.parser.parse(self._expression)
        self.tag_list = None
        return result  # type: ignore
