#
# Copyright (c) 2024, HILSTER - https://hilster.io
# All rights reserved.
#

from typing import Callable, Optional, List

import termcolor  # type: ignore
import os
import sys


def supports_color() -> bool:
    """
    Return ``True`` if the running system's terminal supports color and colors are wanted,
    and ``False`` otherwise.
    """

    if "HTF_NO_COLORS" in os.environ:
        # Colors are disabled by environment variable
        return False

    if "PYCHARM_HOSTED" in os.environ.keys():
        # PyCharm supports colors
        return True

    supported_platform = sys.platform in [
        "linux",
        "linux2",
        "win32",
        "cygwin",
        "darwin",
    ]

    is_a_tty = hasattr(sys.stdout, "isatty") and sys.stdout.isatty()

    return supported_platform and is_a_tty


# instead of cache
color_is_supported = supports_color()


def make_style(
    color: Optional[str] = None,
    on_color: Optional[str] = None,
    attrs: Optional[List[str]] = None,
) -> Callable[[str], str]:
    """
    Return a callable that has a parameter text that will be colored
    according to the parameters if coloring is supported.

    Args:
        color=None: the color for the text
        on_color=None: the background color
        attrs=None: a list of attributes
    """

    if color_is_supported is True:
        return lambda text: termcolor.colored(text=text, color=color, on_color=on_color, attrs=attrs)  # type: ignore
    else:
        return lambda text: text
