#
# Copyright (c) 2024, HILSTER - https://hilster.io
# All rights reserved.
#

from typing import Optional, List
import os
import requests
import re
from dataclasses import dataclass

from htf.termcolors import warning, info


@dataclass(repr=False)
class Version:  # pragma: no cover
    major: Optional[int]
    minor: Optional[int]
    patch: Optional[int]
    postfix: Optional[str]

    @classmethod
    def from_string(cls, version: str) -> "Version":
        parts = list(map(int, version.split(".")[:3]))
        major: Optional[int] = None
        minor: Optional[int] = None
        patch: Optional[int] = None
        postfix: Optional[str] = None

        if len(parts) > 0:
            major = parts[0]

        if len(parts) > 1:
            minor = parts[1]

        if len(parts) > 2:
            patch = parts[2]

        try:
            postfix = version.split(".")[3]
        except IndexError:
            pass

        return Version(major, minor, patch, postfix)

    def __lt__(self, other: "Version") -> bool:
        # make versions sortable
        return [self.major, self.minor, self.patch] < [other.major, other.minor, other.patch]

    def __str__(self) -> str:
        version_str = f"{self.major}"
        if self.minor is not None:
            version_str += f".{self.minor}"
        if self.patch is not None:
            version_str += f".{self.patch}"
        if self.postfix is not None:
            version_str += f".{self.postfix}"
        return version_str

    __repr__ = __str__


def get_htf_versions() -> List[Version]:  # pragma: no cover
    response = requests.get("https://pypi.hilster.io/simple/htf/", timeout=1)
    assert response.status_code == 200
    html = response.text

    matches = re.findall(r'href="/packages/(htf-(\d+\.\d+\.\d+)-[^"]+)"', html)
    # print(f"{matches=}")

    string_versions = set([match[-1] for match in matches])
    versions = [Version.from_string(version) for version in string_versions]

    return sorted(versions)


def get_htf_validated_versions() -> List[Version]:  # pragma: no cover
    response = requests.get("https://pypi.hilster.io/simple/htf-validated/", timeout=1)
    assert response.status_code == 200
    html = response.text

    matches = re.findall(r'href="/packages/(htf_validated-(\d+\.\d+)-[^"]+)"', html)
    # print(f"{matches=}")

    string_versions = set([match[-1] for match in matches])
    versions = [Version.from_string(version) for version in string_versions]

    return sorted(versions)


def check_for_updates() -> None:  # pragma: no cover
    """
    This method checks for updates of htf and htf-validated.
    """

    if os.getenv("HTF_DISABLE_UPDATE_CHECK", None) is not None:
        return

    try:
        import htf

        if htf.core.is_htf_validated_installed():
            latest_version: Version = get_htf_validated_versions()[-1]
            import htf.validation  # type: ignore

            current_version: Version = Version.from_string(htf.validation.__version__)
            product = "htf-validated"
            link = "https://docs.hilster.io/htf/latest/installation.html#installing-a-validated-version-of-htf"
        else:
            latest_version: Version = get_htf_versions()[-1]  # type: ignore
            current_version: Version = Version.from_string(htf.__version__)  # type: ignore
            product = "htf"
            link = "https://docs.hilster.io/htf/latest/installation.html#upgrading"  # type: ignore

        if current_version < latest_version:
            print()
            print(info("Checking for updates"))
            print(
                warning(
                    f"Warning: '{product}-{latest_version}' is available but you are using "
                    f"'{product}-{current_version}'. Please update your {product} installation! "
                    f"See {link} for more details."
                )
            )
    except Exception:
        pass
