#
# Copyright (c) 2024, HILSTER - https://hilster.io
# All rights reserved.
#

from typing import Callable, Any, Generator, Iterable, Optional, List, Union
import itertools
import json
from collections import OrderedDict

import yaml
import csv


def data(iterable: Iterable, unpack: bool = False) -> Callable[[Callable[..., Any]], Any]:
    """
    Create a data driven test using data.

    Args:
        iterable: an iterable used as parameter data source. Can also be a callable.
        unpack=False: if set to ``True`` iterables are unpacked.
    """

    def wrapper(func: Callable[..., Any]) -> Callable[..., Any]:
        # call _test_generator if callable
        if callable(iterable):
            it = iterable()
        else:
            it = iterable

        if hasattr(func, "_test_generator"):
            raise Exception("Data decorators can only be used single.")

        func._test_generator = it  # type: ignore
        func._test_generator_unpack = unpack  # type: ignore

        return func

    return wrapper


class JSONFileIterator:
    """
    JSON file iterator.

    Args:
        filename: the filename to be read.

    Returns:
        iterator: iterator for the file contents.
    """

    def __init__(self, filename: str) -> None:
        self.filename = filename

    def __iter__(self) -> Generator[Any, None, None]:
        with open(self.filename, "r") as fp:
            data = json.load(fp)
        return iter(data)  # type: ignore


def json_data(filename: str, unpack: bool = False) -> Callable[[Callable[..., Any]], Any]:
    """
    Create a data driven test using data from a json file.

    Args:
        filename: the filename to be read.
        unpack=False: if set to ``True`` iterables are unpacked.
    """

    def wrapper(func: Callable[..., Any]) -> Callable[..., Any]:
        func._test_generator = JSONFileIterator(filename)  # type: ignore
        func._test_generator_unpack = unpack  # type: ignore
        return func

    return wrapper


class YAMLFileIterator:
    """
    YAML file iterator.

    Args:
        filename: the filename to be read.

    Returns:
        iterator: iterator for the file contents.
    """

    def __init__(self, filename: str) -> None:
        self.filename = filename

    def __iter__(self) -> Generator[Any, None, None]:
        with open(self.filename, "r") as fp:
            data = yaml.load(fp, Loader=yaml.FullLoader)
        return iter(data)  # type: ignore


def yaml_data(filename: str, unpack: bool = False) -> Callable[[Callable[..., Any]], Any]:
    """
    Create a data driven test using data from a YAML file.

    Args:
        filename: the filename to be read.
        unpack=False: if set to ``True`` iterables are unpacked.
    """

    def wrapper(func: Callable[..., Any]) -> Callable[..., Any]:
        func._test_generator = YAMLFileIterator(filename)  # type: ignore
        func._test_generator_unpack = unpack  # type: ignore
        return func

    return wrapper


class CSVFileIterator:
    """
    CSV file iterator.

    Args:
        filename: the filename to be read.

    Returns:
        iterator: iterator for the file contents as ``dict``.
        convert_to: data types to convert read values to (e.g. float, int)
    """

    def __init__(
        self,
        filename: str,
        convert_to: Optional[Union[Iterable[Any], Generator[Any, None, None], List[Any], Any]] = None,
    ) -> None:
        self.filename = filename
        self.convert_to = convert_to
        try:
            iter(convert_to)  # type: ignore  # check if parameter is iterable
            self.convert_to = convert_to
        except TypeError:
            if callable(convert_to):
                self.convert_to = [convert_to]

    def __iter__(self) -> Generator[Any, None, None]:
        with open(self.filename, "r") as fp:
            reader = csv.reader(fp)
            data = list(reader)
            data_dict = []
            for line in data[1:]:
                d = OrderedDict()
                for index, value in enumerate(line):
                    d[data[0][index]] = value
                data_dict.append(d)
            data = data_dict  # type: ignore

        if self.convert_to is not None:
            if len(self.convert_to) == 1:  # type: ignore
                self.convert_to = self.convert_to * len(data[0])  # type: ignore

            converted = []
            for line in data:
                converted_dict = OrderedDict()
                for index, (name, value) in enumerate(line.items()):  # type: ignore
                    try:
                        converted_value = self.convert_to[index](value)  # type: ignore
                        converted_dict[name] = converted_value
                    except (ValueError, IndexError):
                        converted_dict[name] = value
                converted.append(converted_dict)
            data = converted  # type: ignore

        return iter(data)  # type: ignore


def csv_data(
    filename: str,
    convert_to: Optional[Union[Iterable[Any], Generator[Any, None, None], List[Any], Any]] = None,
) -> Callable[[Callable[..., Any]], Any]:
    """
    Create a data driven test using data from a CSV file.
    The first line of the CSV data is a header. Each column from the first
    is used as the parameter name of the decorated method.
    Each line is supplied as a ``dict``.

    Args:
        filename: the filename to be read.
        convert_to: data types to convert read values to (e.g. float, int)
    """

    def wrapper(func: Callable[..., Any]) -> Callable[..., Any]:
        func._test_generator = CSVFileIterator(filename, convert_to=convert_to)  # type: ignore
        func._test_generator_unpack = False  # type: ignore
        return func

    return wrapper


def product(*iterables: Iterable, **repeat: int) -> Callable[[Callable[..., Any]], Any]:
    """
    Create a data driven test using :func:`itertools.product`
    with a cartesian product of input ``*iterables``.

    Args:
        *iterables: iterables to be passed to :func:`itertools.product`
        repeat=1: number of repetitions.
    """
    return data(itertools.product(*iterables, **repeat), unpack=True)


def permutations(iterable: Iterable, r: Optional[int] = None) -> Callable[[Callable[..., Any]], Any]:
    """
    Create a data driven test using :func:`itertools.permutations`
    with successive ``r`` length permutations of elements in the ``iterable``.

    Args:
        iterable: iterable to be passed to :func:`itertools.permutations`
        r=None: length of permutations
    """
    return data(itertools.permutations(iterable, r=r), unpack=True)


def combinations(iterable: Iterable, r: int) -> Callable[[Callable[..., Any]], Any]:
    """
    Create a data driven test using :func:`itertools.combinations`
    with successive ``r`` length subsequences of elements from ``iterable``.

    Args:
        iterable: iterable to be passed to :func:`itertools.combinations`
        r: length of combinations
    """
    return data(itertools.combinations(iterable, r=r), unpack=True)


def combinations_with_replacement(iterable: Iterable, r: int) -> Callable[[Callable[..., Any]], Any]:
    """
    Create a data driven test using :func:`itertools.combinations`
    with successive ``r`` length subsequences of elements from ``iterable``
    allowing individual elements to be repeated more than once.

    Args:
        iterable: iterable to be passed to :func:`itertools.combinations_with_replacement`
        r: length of combinations
    """
    return data(itertools.combinations_with_replacement(iterable, r=r), unpack=True)
